#ifndef UTSTRUCTLEARN_H_APR_12_2005
#define UTSTRUCTLEARN_H_APR_12_2005

#include <string.h>
#include <stdlib.h>
#include <time.h> //for seeding the random num generator
#include <set>
#include <math.h>
#include "clausefactory.h"
#include "mln.h"
#include "pseudologlikelihood.h"
#include "lbfgsb.h"
#include "ut-mrf.h"
#include "structlearn.h"

class ExistFormula;

class UtStructLearn
{
 public:
  //If there are more than one domain, there must be a one-to-one 
  //correspondence among the clauses in mlns.
  UtStructLearn(Array<MLN*>* const & mlns, const bool& startFromEmptyMLN,
		const string& outMLNFileName, Array<Domain*>* const & domains,
		const Array<string>* const & nonEvidPredNames, 
		const int& maxVars, const bool& cacheClauses, 
		const double& maxCacheSizeMB,
		const bool& tryAllFlips, const bool& sampleClauses, 
		const double& delta, const double& epsilon, 
		const int& minClauseSamples, const int& maxClauseSamples,
		const bool& hasPrior, const double& priorMean, 
		const double& priorStdDev, const bool& wtPredsEqually, 
		const int& lbMaxIter, const double& lbConvThresh,
		const int& looseMaxIter, const double& looseConvThresh, 
		const int& beamSize, const int& bestGainUnchangedLimit,
		const int& numEstBestClauses, 
		const double& minWt, const double& penalty,
		const bool& sampleGndPreds, const double& fraction, 
		const int& minGndPredSamples, const int& maxGndPredSamples,
		const bool& reEvaluateBestCandidatesWithTightParams)
    : mln0_((*mlns)[0]), mlns_(mlns), startFromEmptyMLN_(startFromEmptyMLN), 
      outMLNFileName_(outMLNFileName), domains_(domains), 
      preds_(new Array<Predicate*>), areNonEvidPreds_(new Array<bool>),
      clauseFactory_(new ClauseFactory(maxVars, (*domains_)[0])), 
      cacheClauses_(cacheClauses), origCacheClauses_(cacheClauses), 
      cachedClauses_((cacheClauses) ? (new ClauseHashArray) : NULL), 
      cacheSizeMB_(0), maxCacheSizeMB_(maxCacheSizeMB), 
      tryAllFlips_(tryAllFlips), 
      sampleClauses_(sampleClauses), origSampleClauses_(sampleClauses),
      pll_(NULL), hasPrior_(hasPrior), priorMean_(priorMean), 
      priorStdDev_(priorStdDev), wtPredsEqually_(wtPredsEqually), 
      lbfgsb_(NULL), lbMaxIter_(lbMaxIter), lbConvThresh_(lbConvThresh), 
      looseMaxIter_(looseMaxIter), looseConvThresh_(looseConvThresh),
      beamSize_(beamSize), bestGainUnchangedLimit_(bestGainUnchangedLimit),
      numEstBestClauses_(numEstBestClauses), 
      minGain_(0), minWt_(minWt), penalty_(penalty), 
      sampleGndPreds_(sampleGndPreds), fraction_(fraction), 
      minGndPredSamples_(minGndPredSamples), 
      maxGndPredSamples_(maxGndPredSamples),
      reEvalBestCandsWithTightParams_(reEvaluateBestCandidatesWithTightParams),
    candCnt_(0), iter_(-1), bsiter_(-1), startSec_(-1), indexTrans_(NULL)
    { 
      assert(minWt_ >= 0);
      assert(domains_->size() == mlns_->size());

    srand(time(NULL));
    
    areNonEvidPreds_->growToSize((*domains_)[0]->getNumPredicates(), false);
    for (int i = 0; i < nonEvidPredNames->size(); i++)
    {
      int predId=(*domains_)[0]->getPredicateId((*nonEvidPredNames)[i].c_str());
      if (predId < 0)
      {
        cout << "ERROR: in UtStructLearn::UtStructLearn(). Predicate " 
             << (*nonEvidPredNames)[i] << " undefined." << endl;
        exit(-1);
      }
      (*areNonEvidPreds_)[predId] = true;
    }
    
    (*domains_)[0]->createPredicates(preds_, true);

    if (origSampleClauses_)
    {
      ClauseSampler* cs = new ClauseSampler(delta, epsilon, minClauseSamples, 
                                            maxClauseSamples);
      Clause::setClauseSampler(cs);
      for (int i = 0; i < domains_->size(); i++)
        (*domains_)[i]->newTrueFalseGroundingsStore();
    }    
  }

  
  ~UtStructLearn() 
  {
    
    if (pll_) delete pll_;
    if (lbfgsb_) delete lbfgsb_;
    
    preds_->deleteItemsAndClear();
    delete preds_;
    delete areNonEvidPreds_;
    //delete clauseFactory_;
    
    if (cachedClauses_)
    {
      cachedClauses_->deleteItemsAndClear();
      delete cachedClauses_;
    }
    
    
    //if (origSampleClauses_) delete Clause::getClauseSampler();
    if (indexTrans_) delete indexTrans_;
    
  }

  void diagnoseOnceForAllRun(bool doRelPath, int pNumIters, double pPercentageVotes){
    cout << "Running diagnose once for all  run.." << endl;

    int numCandidatesConsidered = 0;

    if (doRelPath) cout << "Using relational pathfinding\n";
    cout << "Will perform " << pNumIters << " iterations of diagnosis/relearning over the predicates\n";

    //0000000000000000 Doing the preparations: auxiliary data etc 00000000000000000//
    startSec_ = timer_.time();
    bool needIndexTrans =IndexTranslator::needIndexTranslator(*mlns_,*domains_);
    
    //add unit clauses to the MLN
    cout << "adding unit clauses to MLN..." << endl << endl;
    addUnitClausesToMLNs();

    //create auxiliary data for each clause in MLN
    for (int i = 0; i < mln0_->getNumClauses(); i++){
      Clause* c = (Clause*) mln0_->getClause(i);
      c->newAuxClauseData(); 
      //only clauses that are not in any exist. quant. formula's CNF may need 
      //to be translated across domains
      if (isModifiableClause(i)) c->trackConstants();
    }

    //create PseudoLogLikelihood and LBFGSB
    indexTrans_ = (needIndexTrans)? new IndexTranslator(mlns_, domains_) : NULL;
    if (indexTrans_) 
      cout << "The weights of clauses in the CNFs of existential formulas wiil "
           << "be tied" << endl;
    
    pll_ = new PseudoLogLikelihood(areNonEvidPreds_, domains_, wtPredsEqually_, 
                                   sampleGndPreds_, fraction_, 
                                   minGndPredSamples_, maxGndPredSamples_);
    pll_->setIndexTranslator(indexTrans_);
    
    int numClausesFormulas = getNumClausesFormulas();
    lbfgsb_ = new LBFGSB(-1, -1, pll_, numClausesFormulas);
    
    useTightParams();
    
    //compute initial counts for clauses in all MLNs
    cout << "computing counts for initial MLN clauses..." << endl;
    double begSec = timer_.time();
    pllComputeCountsForInitialMLN();
    cout << "computing counts for initial MLN clauses took "; 
    timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
    
    //learn initial wt/score of MLN, and set MLN clauses to learned weights
    cout << "learning the initial weights and score of MLN..." << endl << endl;
    begSec = timer_.time();
    double score;
    Array<double> wts;
    if (!learnAndSetMLNWeights(score)) return; 
    printMLNClausesWithWeightsAndScore(score, -1);

    cout << "learning the initial weights and score of MLN took ";
    timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
    
    //add unit clauses with diff combinations of variables
    cout <<"trying to add unit clause with diff variable combinations to MLN..."
         << endl << endl;
    begSec = timer_.time();
    appendUnitClausesWithDiffCombOfVar(score); // score is updated
    printMLNClausesWithWeightsAndScore(score, -1);
    cout <<"adding unit clause with diff variable combinations to MLN took ";
    timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
   

    //for each clause in MLN, try all sign flips and keep best one in MLN
    if (tryAllFlips_){
      cout << "trying to flip the senses of MLN clauses..." << endl << endl;
      begSec = timer_.time();
      flipMLNClauses(score); 
      printMLNClausesWithWeightsAndScore(score, -1); // score is updated
      cout << "trying to flip the senses of MLN clauses took ";;
      timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
    }

    iter_ = -1;
    //first prune clauses--we don't want crappy clauses to interfere with the search
    pruneMLN(score);
    
    cout << "After initial pruning the mln is;\n";
    printMLNClausesWithWeightsAndScore(score, -1);

    //here we want to keep the pathfinding clauses out of the beam;
    //that's why we are using this array
    Array<Clause*> trulyInitialClauses; //need to clean them up afterwards   
    for (int i = 0; i < mln0_->getNumClauses(); i++) {
      Clause* c = (Clause*) mln0_->getClause(i);
      c->getAuxClauseData()->reset();
      if (isModifiableClause(i)) {
		c->trackConstants();
		trulyInitialClauses.append(new Clause(*c));
      }
    }

    Array<Clause*> initialClauses;
    Array<Clause*> bestCandidates;

    if (doRelPath) {
      //DO RELATIONAL PATHFINDING    
      //set the penalty and weight to 0 for this part. 
      double oldPenalty = penalty_;
      penalty_ = 0;
      double oldMinWt = minWt_;
      minWt_ = 0;
    
      Array<Clause*> bestRelPathCandidates;
      ClauseOpHashArray generatedClauses;
      performRelationalPathfindingForAll((*domains_)[0], generatedClauses);
    
      cout << "Relational pathfinding found the following clauses\n";
      for (int i = 0; i < generatedClauses.size(); i++) {
		initialClauses.append(generatedClauses[i]);
		generatedClauses[i]->print(cout, (*domains_)[0]); cout << endl;
      }
      cout << "Done printing clauses found by relational pathfinding\n";
    
      numCandidatesConsidered += pointedBeamSearch(initialClauses, score, bestRelPathCandidates, true, true); 
   
      if (appendToAndRemoveFromMLNs(bestRelPathCandidates, score))
		cout << "Managed to append some clause to the MLN\n";
      else {
		cout << "Rel pathfinding didn't find anything worthwhile.\n";
      }
    
      penalty_ = oldPenalty;
      minWt_ = oldMinWt;
      //DONE WITH REL PATHFINDING
    }
    cout << "After pathfinding, the mln is:\n";
    printMLNClausesWithWeightsAndScore(score, -1);
    //000000000000000000000000000 Done with the initial preparations 000000000000000//

    //set up various params for inference:
    cout << "Preparing the gibbs and walksat parameters..." << endl;
    GibbsParams gibbsParams;
    MaxWalksatParams * wsparams = new MaxWalksatParams;
    prepareGibbsAndMWSParams(gibbsParams, wsparams);
        
    int randomDomain = rand() % domains_->size(); 
    cout << "Randomly picked domain " << randomDomain << endl;;
    Domain * firstInList = (*domains_)[randomDomain];  
    //Domain * firstInList = (*domains_)[0];  

    int numPreds = firstInList->getNumPredicates();
    
    //Outer-most loop. How many times do we want to go over all the
    //predicates
    for (int ii = 0; ii < pNumIters; ii++) { 
      bool sawChange = false;
            
      Array<UTMRF *> mrfs; //collect them for deletion
      ClauseOpHashArray clausesForAdd;
      ClauseOpHashArray clausesForRemove;
      ClauseOpHashArray clausesForPathFind;

      //cycle through each predicate
      for(int p = 0; p < numPreds; p++ ){  
		const char * nameOfCurrPred = firstInList->getPredicateName(p);
	
		if (strstr(nameOfCurrPred, "SK")) continue;   //equals predicates
		cout << "Working on predicate " << nameOfCurrPred << endl; 
	
		UTMRF * theMRF = runDiagnosticInference(gibbsParams, wsparams,firstInList,
						nameOfCurrPred, p, trulyInitialClauses);
		mrfs.append(theMRF);
	
		Array<ClauseWithInfo*>* theClausesWithCounts = theMRF->getUngndClausesWithInfo();
	
		cout << "printing the clauses with counts\n";
		for (int i = 0; i < theClausesWithCounts->size(); i++){
	  		ClauseWithInfo * currClauseWI = (*theClausesWithCounts)[i];
	  		cout << "__________________________________\n";
	  		currClauseWI->theClause->print(cout, firstInList); cout << endl;
	  		for (int j = 0; j < currClauseWI->theInfo.size(); j++){
	    		currClauseWI->theInfo[j]->printCounts();
	  		}
		}
       
      	int indexOfNextOne = -1;
		int doWhat = -1;  //generalize or specialize
		int positionInInfo = -1;
		
		while (true) {
		  double numBest = nextMostUrgent(theClausesWithCounts, indexOfNextOne,
						  doWhat, positionInInfo);
		  if (numBest <= 0) {
		    cout << "No more clauses need revision\n";
		    break;
		  }
		  
		  assert(indexOfNextOne >= 0);
		  assert(positionInInfo >= 0);
		  
		  //this is the clause that needs updating right now
		  ClauseWithInfo * theCwi = (*theClausesWithCounts)[indexOfNextOne];
		  //int positionInClause = theCwi->theInfo[positionInInfo]->predLocation;
		  
		  cout << "picked clause " ; 
		  theCwi->theClause->print(cout, firstInList); cout << endl;
	
		  if (numBest <= pPercentageVotes * theCwi->theInfo[positionInInfo]->getTotals()){
		    cout << "The number of votes for best revision is not large enough..." << endl;
		    cout << "The actual percentage of votes is "<< pPercentageVotes * theCwi->theInfo[positionInInfo]->getTotals() << endl;
		    cout << "numbest is " << numBest << endl;
	
		    break; 
		  }

		  cout << "doWhat is " << doWhat << endl;
		  
		  assert(doWhat == UTMRF::GENERALIZE || doWhat == UTMRF::SPECIALIZE);
		  
		  //set them all to 0. We've already included this clause
		  for (int i = 0; i < theCwi->theInfo.size(); i++) {
		    theCwi->theInfo[i]->clauseClassCounts[doWhat] = 0;
		  }
	  	  
		  if (doWhat == UTMRF::GENERALIZE){
		    clausesForRemove.append(theCwi->theClause);
		    cout << "Added the clause in the clausesForRemove \n";
		  }
		  else {//must be specialized
		    clausesForAdd.append(theCwi->theClause);
		    cout << "Added the clause in the clausesForAdd \n";
		  }
		} //while true
      }

      useLooseParams();

      cout << "+++++++++++++++++++++++++++++++++++++++++++++++++\n";
      cout << "The following clauses are considered for remove\n";
      for (int i = 0; i < clausesForRemove.size(); i++) {
		clausesForRemove[i]->print(cout, firstInList); cout << endl;
      }
      cout << "Done with clauses for remove\n";

      cout << "The following clauses are considered for add\n";
      for (int i = 0; i < clausesForAdd.size(); i++) {
		clausesForAdd[i]->print(cout, firstInList); cout << endl;
	  }
      cout << "Done with clauses for add\n";
      cout << "+++++++++++++++++++++++++++++++++++++++++++++++++\n";
  
      Array<Clause *> initClauses;
      for (int i = 0; i < clausesForRemove.size(); i++)
		initClauses.append(clausesForRemove[i]);

      Array<Clause *> bestClauses;      
      if (initClauses.size() > 0)
      	numCandidatesConsidered +=  pointedBeamSearch(initClauses, score, bestClauses, false);
      
      bool ok = findBestAndAddToMLN(bestClauses, score);

      initClauses.clear();
      bestClauses.clear();
      for (int i = 0; i < clausesForAdd.size(); i++) 
		initClauses.append(clausesForAdd[i]);
      
      if (initClauses.size() > 0)
      	numCandidatesConsidered += pointedBeamSearch(initClauses, score, bestClauses, true);
      
      ok = findBestAndAddToMLN(bestClauses, score);

      useTightParams();
      
      if (!sawChange && ok) sawChange = true;	   
     	
      for (int i = 0; i < mrfs.size(); i++)
		delete mrfs[i];

      if (!sawChange) break;
    } 

    useTightParams();
    
    //Prune each non-unit clause in MLN if it does not decrease score    
    cout << "pruning clauses from MLN..." << endl << endl;
    begSec = timer_.time();
    pruneMLN(score);
    cout << "pruning clauses from MLN took ";
    timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
    
    printMLNClausesWithWeightsAndScore(score, -1);
    printMLNToFile(NULL, -2);
    
    cout << "time taken for structure learning = ";
    timer_.printTime(cout, timer_.time()-startSec_); cout << endl << endl;

    cout << "Number of candidates considered was " << numCandidatesConsidered << endl;

  }//diagnoseOnceForAllRun()

//Added by Tuyen
double reviseOneClause(int pNumIters, double pPercentageVotes){
    cout << "Running reviseOneClause  run.." << endl;

    int numCandidatesConsidered = 0;

    startSec_ = timer_.time();
    bool needIndexTrans =IndexTranslator::needIndexTranslator(*mlns_,*domains_);
    
    //add unit clauses to the MLN
    cout << "adding unit clauses to MLN..." << endl << endl;
    addUnitClausesToMLNs();

    //create auxiliary data for each clause in MLN
    for (int i = 0; i < mln0_->getNumClauses(); i++){
      Clause* c = (Clause*) mln0_->getClause(i);
      c->newAuxClauseData(); 
      //only clauses that are not in any exist. quant. formula's CNF may need 
      //to be translated across domains
      if (isModifiableClause(i)) c->trackConstants();
    }

    //create PseudoLogLikelihood and LBFGSB
    indexTrans_ = (needIndexTrans)? new IndexTranslator(mlns_, domains_) : NULL;
    if (indexTrans_) 
      cout << "The weights of clauses in the CNFs of existential formulas wiil "
           << "be tied" << endl;
    
    pll_ = new PseudoLogLikelihood(areNonEvidPreds_, domains_, wtPredsEqually_, 
                                   sampleGndPreds_, fraction_, 
                                   minGndPredSamples_, maxGndPredSamples_);
    pll_->setIndexTranslator(indexTrans_);
    
    int numClausesFormulas = getNumClausesFormulas();
    lbfgsb_ = new LBFGSB(-1, -1, pll_, numClausesFormulas);
    
    useTightParams();
    
    //compute initial counts for clauses in all MLNs
    cout << "computing counts for initial MLN clauses..." << endl;
    double begSec = timer_.time();
    pllComputeCountsForInitialMLN();
    cout << "computing counts for initial MLN clauses took "; 
    timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
    
    //learn initial wt/score of MLN, and set MLN clauses to learned weights
    cout << "learning the initial weights and score of MLN..." << endl << endl;
    begSec = timer_.time();
    double score = -1E+10;
    Array<double> wts;
    if (!learnAndSetMLNWeights(score)) return score; 
    printMLNClausesWithWeightsAndScore(score, -1);

    cout << "learning the initial weights and score of MLN took ";
    timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
    
    
    //add unit clauses with diff combinations of variables
    cout <<"trying to add unit clause with diff variable combinations to MLN..."
         << endl << endl;
    begSec = timer_.time();
    appendUnitClausesWithDiffCombOfVar(score); // score is updated
    printMLNClausesWithWeightsAndScore(score, -1);
    cout <<"adding unit clause with diff variable combinations to MLN took ";
    timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
    
   

    //for each clause in MLN, try all sign flips and keep best one in MLN
    if (tryAllFlips_){
      cout << "trying to flip the senses of MLN clauses..." << endl << endl;
      begSec = timer_.time();
      flipMLNClauses(score); 
      printMLNClausesWithWeightsAndScore(score, -1); // score is updated
      cout << "trying to flip the senses of MLN clauses took ";;
      timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
    }

    iter_ = -1;
    //first prune clauses--we don't want crappy clauses to interfere with the search
    pruneMLN(score);
    
    cout << "After initial pruning the mln is;\n";
    printMLNClausesWithWeightsAndScore(score, -1);

    //here we want to keep the pathfinding clauses out of the beam;
    //that's why we are using this array
    Array<Clause*> trulyInitialClauses; //need to clean them up afterwards   
    for (int i = 0; i < mln0_->getNumClauses(); i++) {
      Clause* c = (Clause*) mln0_->getClause(i);
      c->getAuxClauseData()->reset();
      if (isModifiableClause(i)) {
		c->trackConstants();
		trulyInitialClauses.append(new Clause(*c));
      }
    }

    Array<Clause*> initialClauses;
    Array<Clause*> bestCandidates;

    //set up various params for inference:
    cout << "Preparing the gibbs and walksat parameters..." << endl;
    GibbsParams gibbsParams;
    MaxWalksatParams * wsparams = new MaxWalksatParams;
    prepareGibbsAndMWSParams(gibbsParams, wsparams);
        
    int randomDomain = rand() % domains_->size(); 
    cout << "Randomly picked domain " << randomDomain << endl;;
    Domain * firstInList = (*domains_)[randomDomain];  
    //Domain * firstInList = (*domains_)[0];  

    int numPreds = firstInList->getNumPredicates();
    
    //Outer-most loop. How many times do we want to go over all the
    //predicates
    for (int ii = 0; ii < pNumIters; ii++) { 
      bool sawChange = false;
            
      Array<UTMRF *> mrfs; //collect them for deletion
      ClauseOpHashArray clausesForAdd;
      ClauseOpHashArray clausesForRemove;
      ClauseOpHashArray clausesForPathFind;

      //cycle through each predicate
      for(int p = 0; p < numPreds; p++ ){  
		const char * nameOfCurrPred = firstInList->getPredicateName(p);
	
		if (strstr(nameOfCurrPred, "SK")) continue;   //equals predicates
		cout << "Working on predicate " << nameOfCurrPred << endl; 
	
		UTMRF * theMRF = runDiagnosticInference(gibbsParams, wsparams,firstInList,
						nameOfCurrPred, p, trulyInitialClauses);
		mrfs.append(theMRF);
	
		Array<ClauseWithInfo*>* theClausesWithCounts = theMRF->getUngndClausesWithInfo();
	
		cout << "printing the clauses with counts\n";
		for (int i = 0; i < theClausesWithCounts->size(); i++){
	  		ClauseWithInfo * currClauseWI = (*theClausesWithCounts)[i];
	  		cout << "__________________________________\n";
	  		currClauseWI->theClause->print(cout, firstInList); cout << endl;
	  		for (int j = 0; j < currClauseWI->theInfo.size(); j++){
	    		currClauseWI->theInfo[j]->printCounts();
	  		}
		}
       
      	int indexOfNextOne = -1;
		int doWhat = -1;  //generalize or specialize
		int positionInInfo = -1;
		
		while (true) {
		  double numBest = nextMostUrgent(theClausesWithCounts, indexOfNextOne,
						  doWhat, positionInInfo);
		  if (numBest <= 0) {
		    cout << "No more clauses need revision\n";
		    break;
		  }
		  
		  assert(indexOfNextOne >= 0);
		  assert(positionInInfo >= 0);
		  
		  //this is the clause that needs updating right now
		  ClauseWithInfo * theCwi = (*theClausesWithCounts)[indexOfNextOne];
		  //int positionInClause = theCwi->theInfo[positionInInfo]->predLocation;
		  
		  cout << "picked clause " ; 
		  theCwi->theClause->print(cout, firstInList); cout << endl;
	
		  if (numBest <= pPercentageVotes * theCwi->theInfo[positionInInfo]->getTotals()){
		    cout << "The number of votes for best revision is not large enough..." << endl;
		    cout << "The actual percentage of votes is "<< pPercentageVotes * theCwi->theInfo[positionInInfo]->getTotals() << endl;
		    cout << "numbest is " << numBest << endl;
	
		    break; 
		  }

		  cout << "doWhat is " << doWhat << endl;
		  
		  assert(doWhat == UTMRF::GENERALIZE || doWhat == UTMRF::SPECIALIZE);
		  
		  //set them all to 0. We've already included this clause
		  for (int i = 0; i < theCwi->theInfo.size(); i++) {
		    theCwi->theInfo[i]->clauseClassCounts[doWhat] = 0;
		  }
	  	  
		  if (doWhat == UTMRF::GENERALIZE){
		    clausesForRemove.append(theCwi->theClause);
		    cout << "Added the clause in the clausesForRemove \n";
		  }
		  else {//must be specialized
		    clausesForAdd.append(theCwi->theClause);
		    cout << "Added the clause in the clausesForAdd \n";
		  }
		} //while true
      }

      useLooseParams();

      cout << "+++++++++++++++++++++++++++++++++++++++++++++++++\n";
      cout << "The following clauses are considered for remove\n";
      for (int i = 0; i < clausesForRemove.size(); i++) {
		clausesForRemove[i]->print(cout, firstInList); cout << endl;
      }
      cout << "Done with clauses for remove\n";

      cout << "The following clauses are considered for add\n";
      for (int i = 0; i < clausesForAdd.size(); i++) {
		clausesForAdd[i]->print(cout, firstInList); cout << endl;
	  }
      cout << "Done with clauses for add\n";
      cout << "+++++++++++++++++++++++++++++++++++++++++++++++++\n";
  
      Array<Clause *> initClauses;
      for (int i = 0; i < clausesForRemove.size(); i++)
		initClauses.append(clausesForRemove[i]);

      Array<Clause *> bestClauses;      
      if (initClauses.size() > 0)
      	numCandidatesConsidered +=  pointedBeamSearch(initClauses, score, bestClauses, false);
      
      bool ok = findBestAndAddToMLN(bestClauses, score);

      initClauses.clear();
      bestClauses.clear();
      for (int i = 0; i < clausesForAdd.size(); i++) 
		initClauses.append(clausesForAdd[i]);
      
      if (initClauses.size() > 0)
      	numCandidatesConsidered += pointedBeamSearch(initClauses, score, bestClauses, true);
      
      ok = findBestAndAddToMLN(bestClauses, score);

      useTightParams();
      
      if (!sawChange && ok) sawChange = true;	   
     	
      for (int i = 0; i < mrfs.size(); i++)
		delete mrfs[i];

      if (!sawChange) break;
    } 

    useTightParams();
    
    //Prune each non-unit clause in MLN if it does not decrease score    
    cout << "pruning clauses from MLN..." << endl << endl;
    begSec = timer_.time();
    pruneMLN(score);
    cout << "pruning clauses from MLN took ";
    timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
    
    printMLNClausesWithWeightsAndScore(score, -1);
    printMLNToFile(NULL, -2);
    
    cout << "time taken for structure learning = ";
    timer_.printTime(cout, timer_.time()-startSec_); cout << endl << endl;

    cout << "Number of candidates considered was " << numCandidatesConsidered << endl;

	return score;
  }//reviseOneClause()


  //zero means nothing needed to be generalized or specialized
  //returns the largest number
  double nextMostUrgent(Array<ClauseWithInfo*>* theClausesWithCounts, 
		      int & index, int & doWhat, int & positionInInfo){
    int bestIndex = -1; //index of clause within the array
    int bestLocInInfo = -1; 
    int bestDoWhat = -1; //generalize or spec.
    double bestCount = 0; //largest number 

    for (int i = 0; i < theClausesWithCounts->size(); i++){ //going over each clause
      ClauseWithInfo* cwi = (*theClausesWithCounts)[i];
      //    double weight = fabs(cwi->theClause->getWt());
      double weight = 1;
      for (int j = 0; j < cwi->theInfo.size(); j++){ //from each perspective
	ClauseInfoWrtPred *cip = cwi->theInfo[j];
     
	if (cip->clauseClassCounts[UTMRF::GENERALIZE] > 
	    cip->clauseClassCounts[UTMRF::SPECIALIZE] ){
	  if (cip->clauseClassCounts[UTMRF::GENERALIZE] * weight > bestCount) {
	    bestCount = cip->clauseClassCounts[UTMRF::GENERALIZE] * weight;
	    bestDoWhat = UTMRF::GENERALIZE;
	    bestIndex = i;
	    bestLocInInfo = j;
	  }
	}else if (cip->clauseClassCounts[UTMRF::SPECIALIZE] * weight > bestCount) {
	  bestCount = cip->clauseClassCounts[UTMRF::SPECIALIZE] * weight;
	  bestDoWhat = UTMRF::SPECIALIZE;
	  bestIndex = i;
	  bestLocInInfo = j;
	}
      }      
    }

    doWhat = bestDoWhat;
    index = bestIndex;
    positionInInfo = bestLocInInfo;
    return bestCount;
  }

  void specialize(ClauseWithInfo * cwi, ClauseOpHashArray & newClauses, int op) {
    addPredicateToClause(cwi->theClause, op, -1, newClauses);
  }

  //try all possible ways of removing a predicate from the clause
  //thrownout collects all clauses that we considered adding but did not add because they were 
  //alreay there
  void generalize(ClauseWithInfo * cwi, int bestPredPosition, 
		  ClauseOpHashArray & newClauses, int op){
    cout << "Now in generalize method\n";
    int numPredicates = cwi->theClause->getNumPredicates();
    cout << "Num Predicates =  " << numPredicates << endl;
    for (int i = 0; i < numPredicates; i++){
      
      if (i != bestPredPosition){
	cout << "The size of newClauses before addition of clause is " << newClauses.size() << endl;
	cout << "going into removeParticularPredicateFromClause\n";
	removeParticularPredicateFromClause(i, cwi->theClause, op, 
					    cwi->theClause->getAuxClauseData()->removedClauseIdx, newClauses);
	cout << "The size of newClauses after addition of clause is " << newClauses.size() << endl;
      }
    }
  } //generalize

 
  void prepareSeenArrays(int numInFirst, int numInSecond, int numConstants,
			 Array<int*>* seenArray1, Array<int *>* seenArray2){
    assert (numInFirst > 0 && numInSecond > 0);
    
    for (int i = 0; i < numInFirst; i++){
      int * seen = new int[numConstants];
      for (int j = 0; j < numConstants; j++) 
	seen[j] = -1;
      seenArray1->append(seen);
    }

    for (int i = 0; i < numInSecond; i++){
      int * seen = new int[numConstants];
      for (int j = 0; j < numConstants; j++) 
	seen[j] = -1;
      seenArray2->append(seen);
    }


  }

  void clearSeenArrays(int numInFirst, int numInSecond, 
		       Array<int*> *seenArray1, Array<int*> *seenArray2){
    for (int i = 0; i < numInFirst; i++)
      delete (*seenArray1)[i];
    for (int i = 0; i < numInSecond; i++)
      delete (*seenArray2)[i];
    delete seenArray1;
    delete seenArray2;
  }


  bool learnAndSetMLNWeights(double& score)
  {
    Array<double> priorMeans, priorStdDevs;
    double tmpScore = score;
    pllSetPriorMeansStdDevs(priorMeans, priorStdDevs, 0, NULL);
    int numClausesFormulas = getNumClausesFormulas();
    Array<double> wts;
    wts.growToSize(numClausesFormulas + 1);
      //indexTrans_ is already created
    int iter; bool error; double elapsedSec;
    tmpScore = maximizeScore(numClausesFormulas, 0, &wts, NULL, NULL,
                             iter, error, elapsedSec);

    cout <<"****The weights learned at this point are :\n";
    for (int i = 0; i < wts.size(); i++)
      cout << wts[i] << " ";
    cout << endl;


    if (error) 
    { 
      cout << "LBFGSB failed to find wts" << endl; 
      return false;
    }
    else 
    { 
      score = tmpScore;
      printNewScore((Clause*)NULL, NULL, iter, elapsedSec, score, 0, 0);
    }
    
    updateWts(wts, NULL, NULL);

    return true;
  }

  //calculates the differences between the probabilities output for
  //the provided groundings and the actual values they need to reach.
  //the predicates in the groundings hasharray will be deleted as a
  //result of calling this method
  void computeProbabilityDifferences(PredicateHashArray * trueGroundings, 
				     PredicateHashArray *falseGroundings, 
				     Domain * currDomain, int predId,
				     double &posDifference, double &negDifference) {

    GroundPredicateHashArray * pseudoQueries = new GroundPredicateHashArray;
    Array<bool>* truthValues = new Array<bool>;
    currDomain->getDB()->setClosedWorld(predId, false);
       
    for (int i = 0; i < trueGroundings->size(); i++) {
      GroundPredicate *gp = new GroundPredicate((*trueGroundings)[i]);
      pseudoQueries->append(gp);
      currDomain->getDB()->setValue((*trueGroundings)[i], UNKNOWN);
      truthValues->append(true);
    }

    for (int i = 0; i < falseGroundings->size(); i++) {
      //      GroundPredicate *gp = new GroundPredicate((*falseGroundings)[i]);
      // pseudoQueries->append(gp);
      currDomain->getDB()->setValue((*falseGroundings)[i], FALSE);
      //truthValues->append(false);
    }

    if (pseudoQueries->size() == 0) {
      cout << "Skipping because there are no true groundings of this predicate\n";
      return;
    }

    UTMRF * forDiagnosing = new UTMRF(pseudoQueries, currDomain, currDomain->getDB(), mln0_);
    forDiagnosing->calculateWeightsOfGroundings();

    assert(pseudoQueries->size() == truthValues->size());

    for (int i = 0; i < pseudoQueries->size(); i++) {
      double thisQueryProb = (*pseudoQueries)[i]->getProbability(0);
      //cout << "The current query is "; 
      //(*pseudoQueries)[i]->print(cout, currDomain);
      //cout << "The probability of this query is" << thisQueryProb << endl; cout.flush(); 
      assert(thisQueryProb >= 0 && thisQueryProb <= 1);
      
      if ((*truthValues)[i] == true) {
	//cout << "This is a true query\n"; 
	posDifference += (1 - thisQueryProb);
      }
      else {
	//cout << "This is a false query\n";
	negDifference += thisQueryProb;
      }
      
      delete (*pseudoQueries)[i];
    }
    //    cout << "Done calculating total difference: " << totalDifference << endl; cout.flush();

    for (int i = 0; i < trueGroundings->size(); i++) {
      currDomain->getDB()->setValue((*trueGroundings)[i], TRUE);
      delete (*trueGroundings)[i];
    }
    for (int i = 0; i < falseGroundings->size(); i++) {
      currDomain->getDB()->setValue((*falseGroundings)[i], FALSE);
      delete (*falseGroundings)[i];
    }
    currDomain->getDB()->setClosedWorld(predId, true);
    
      
    delete pseudoQueries;
    delete forDiagnosing;
    delete truthValues;

    //return totalDifference; 
  }


 /////////////// relational pathfinding stuff ///////
    private:
  void performRelationalPathfindingForAll(Domain * domain, ClauseOpHashArray & newClauses){

    //first prepare the true gndings of all binary non-equality predicates.
    Array<PredicateHashArray *>* trueGndings = new Array<PredicateHashArray *>;
    for (int predId = 0; predId < domain->getNumPredicates(); predId++) {
      if (domain->getPredicateTemplate(predId)->getNumTerms() == 2 &&
	  !domain->getPredicateTemplate(predId)->isEqualPred() &&
	  strncmp(domain->getPredicateName(predId), "same", 4) != 0) {
	PredicateHashArray * thisPredGndings = new PredicateHashArray;
	getTruePredGndings(predId, domain, thisPredGndings);
	trueGndings->append(thisPredGndings);
      }
      else{
	PredicateHashArray * thisPredGndings = NULL;
	trueGndings->append(thisPredGndings);
      }
    }

    for (int predId = 0; predId < domain->getNumPredicates(); predId++) {
      //we are looking for binary, non-equality predicates
      if (domain->getPredicateTemplate(predId)->getNumTerms() == 2 &&
	  !domain->getPredicateTemplate(predId)->isEqualPred() && 
	  strncmp(domain->getPredicateName(predId), "same", 4) != 0) {
	performRelationalPathfindingForPredicate(predId, domain, trueGndings, newClauses);
      }
    }
    
    for (int i = 0; i < trueGndings->size(); i++) 
      if ((*trueGndings)[i] != NULL)
	delete (*trueGndings)[i];
    delete trueGndings;
  }
					     
  void performRelationalPathfindingForPredicate(int predId, Domain * domain, 
						Array<PredicateHashArray*>* trueGndings,
						ClauseOpHashArray & newClauses){

    assert(domain->getPredicateTemplate(predId)->getNumTerms() == 2);

    assert ((*trueGndings)[predId] != NULL);

    // Step 1.  For each true gnding of this predicate...
    
    const PredicateHashArray * trueGndingsOfPred = (*trueGndings)[predId];

    int numTrueGndPreds = trueGndingsOfPred->size();
    if (numTrueGndPreds == 0) {
      cout << "------Predicate with id " << predId << " has not true groundings? \n";
      return;
    }

    for (int predIndex = 0; predIndex < numTrueGndPreds; predIndex++) {
      Predicate * startingPred = (*trueGndingsOfPred)[predIndex];    

      //Done with step 1.

      // Step 2. Look for paths of length 2 by expanding from the constants in starting node.
      int leftConstId = startingPred->getTerm(0)->getId();
      int rightConstId = startingPred->getTerm(1)->getId();

      //this assumes that the constants are numbered from 0 to numConsts - 1
      //for a constant that cooccurs with our const of interest, we will store the predicate in which this happens
      Array <Array<Predicate*>*> leftConsts;
      Array <Array<Predicate*>*> rightConsts;
      
      //these need to be cleaned up afterwards
      leftConsts.growToSize(domain->getNumConstants());
      rightConsts.growToSize(domain->getNumConstants());
      for (int i = 0; i < leftConsts.size(); i++){
	leftConsts[i] = new Array<Predicate *>;
	rightConsts[i] = new Array<Predicate *>;
      }

      //go over the true groundings of all binary non-equality predicates
      for (int p = 0; p < domain->getNumPredicates(); p++) {
	if (domain->getPredicateTemplate(p)->getNumTerms() == 2 &&
	    !domain->getPredicateTemplate(p)->isEqualPred()&& 
	    strncmp(domain->getPredicateName(p), "same", 4) != 0) {

	  assert((*trueGndings)[p] != NULL);

	  PredicateHashArray * currPredTrueGndings = (*trueGndings)[p];
	  for (int tg = 0; tg < currPredTrueGndings->size(); tg++){
	    Predicate *currPredGnding = (*currPredTrueGndings)[tg];
	    
	    int thisPredConst1 = currPredGnding->getTerm(0)->getId();
	    int thisPredConst2 = currPredGnding->getTerm(1)->getId();
	    
	    assert (thisPredConst1 >= 0 && thisPredConst2 >= 0);

	    if (leftConstId == thisPredConst1)
	      leftConsts[thisPredConst2]->append(currPredGnding);
	    else if (leftConstId == thisPredConst2)
	      leftConsts[thisPredConst1]->append(currPredGnding);
	    
	    if (rightConstId == thisPredConst1)
	      rightConsts[thisPredConst2]->append(currPredGnding);
	    else if (rightConstId == thisPredConst2)
	      rightConsts[thisPredConst1]->append(currPredGnding);
	  }
	}
      }
    
      // Step 3. Look for hits and reconstruct the clauses
      for (int i = 0; i < leftConsts.size(); i++){
	if (leftConsts[i]->size() != 0 && rightConsts[i]->size() != 0) {
	  //we have a match
	  for (int j = 0; j < leftConsts[i]->size(); j++) {
	    for (int k = 0; k < rightConsts[i]->size(); k++) {
	      //	      cout << "Found relational path: "; cout.flush();
	      //startingPred->print(cout, domain); cout << " , "; cout.flush();
	      //(*leftConsts[i])[j]->print(cout,domain); cout << " , "; cout.flush();
	      //(*rightConsts[i])[k]->print(cout, domain); cout << endl; cout.flush();
	      
	      Clause * newClause = new Clause();
	      preparePredicateAndAddToClause(newClause, startingPred, true);
	      preparePredicateAndAddToClause(newClause, (*leftConsts[i])[j], false);
	      preparePredicateAndAddToClause(newClause, (*rightConsts[i])[k], false);

	      newClause->canonicalize();
	      AuxClauseData * acd = new AuxClauseData();
	      newClause->setAuxClauseData(acd);
	      newClause->print(cout, domain); cout << endl;
	      
	      newClauses.append(newClause);
	    }
	  }

	}
      }    

      //clean up
      for (int i = 0; i < leftConsts.size(); i++){
	delete leftConsts[i];
	delete rightConsts[i];
      }
    }

  }

  //this method is used by the relational pathfinding to construct the found new clauses
  void preparePredicateAndAddToClause(Clause * clause, Predicate *givenGndPred, bool predSense) {
    Predicate * newPredicate = new Predicate(givenGndPred->getTemplate());
    for (int t = 0; t < givenGndPred->getNumTerms(); t++) {
      int newTermId = (-1) * givenGndPred->getTerm(t)->getId() - 1;
      Term *newTerm = new Term(newTermId, newPredicate, true);
      newPredicate->appendTerm(newTerm);
    }

    newPredicate->setSense(predSense);
    clause->appendPredicate(newPredicate);
  }

   void getTruePredGndings(int predId, Domain * currDomain, 
			  PredicateHashArray * trueGroundings){
    Array <Predicate *> allGndings;
    Predicate::createAllGroundings(predId, currDomain, allGndings);
    
    for (int g = 0; g < allGndings.size(); g++) {
      TruthValue tv = currDomain->getDB()->getValue(allGndings[g]);
      if (tv == TRUE) {
	if (trueGroundings->append(allGndings[g]) < 0)
	  delete allGndings[g];
      }
      else if (tv == FALSE) 
	  delete allGndings[g];
      else //tv must be either true or false
	assert(false);
    }
  }

  //caller is responsible for deleting the contents of the
  //trueGroundigs and falseGroundigns arrays
  void splitAllPredGndings(int predId, Domain * currDomain, 
			   PredicateHashArray * trueGroundings,
			   PredicateHashArray * falseGroundings){
    
    Array <Predicate *> allGndings;
    Predicate::createAllGroundings(predId, currDomain, allGndings);
    
    for (int g = 0; g < allGndings.size(); g++) {
      TruthValue tv = currDomain->getDB()->getValue(allGndings[g]);
      if (tv == TRUE) {
	if (trueGroundings->append(allGndings[g]) < 0)
	  delete allGndings[g];
      }
      else if (tv == FALSE) {
	if (falseGroundings->append(allGndings[g]) < 0)
	  delete allGndings[g];
      }
      else
	assert(false);
    }
  }
 

  //////////////////////////// beam search ////////////////////////////////
 private:
    // priorMeans and priorStdDevs should be one larger than the current
    // num of clauses in the MLN. 
    // The clauses in initClauses must have their AuxClauseData created.
  int beamSearch(const Array<Clause*>& initClauses, 
                  const Array<Clause*>& initMLNClauses, const double& prevScore,
		 Array<Clause*>& bestClauses, bool justUnaryPreds = false)
  {
    int iterBestGainUnchanged = 0; bsiter_ = -1;
    int numCandidatesConsidered = 0;
    Array<double> priorMeans, priorStdDevs;
    ClauseOpHashArray* beam = new ClauseOpHashArray;
    for (int i = 0; i < initClauses.size(); i++) beam->append(initClauses[i]);

    int numClausesFormulas = getNumClausesFormulas();
    Array<double> wts;
    wts.growToSize(numClausesFormulas + 2); // first slot is unused
    Array<double> origWts(numClausesFormulas); // current clause/formula wts
    if (indexTrans_) indexTrans_->getClauseFormulaWts(origWts);
    else             mln0_->getClauseWts(origWts);

      //set the prior means and std deviations
    setPriorMeansStdDevs(priorMeans, priorStdDevs, 1, NULL);
    if (indexTrans_) indexTrans_->appendClauseIdxToClauseFormulaIdxs(1, 1);

    bool error;
    double begIterSec, bsec;
    Array<Clause*> candidates;
    while (!beam->empty() && iterBestGainUnchanged < bestGainUnchangedLimit_)
    {
      begIterSec = timer_.time();
      bsiter_++; 

      //if (bsiter_ == 1) break; // for testing purposes only
      //     cout << endl << "BEAM SEARCH ITERATION " << bsiter_ << endl << endl;

      //      cout << "creating candidate clauses..." << endl;
      candidates.clear();
      bsec = timer_.time();
      
      if (bsiter_==0) createCandidateClauses(beam, candidates, &initMLNClauses, justUnaryPreds);
      else            createCandidateClauses(beam, candidates, NULL, justUnaryPreds);
      
      numCandidatesConsidered += candidates.size();

      cout << "num of candidates created = " << candidates.size() << "; took ";
      timer_.printTime(cout, timer_.time()-bsec); cout << endl << endl;

      cout << "evaluating gain of candidates..." << endl << endl;
      bsec = timer_.time();
      for (int i = 0; i < candidates.size(); i++)
        countAndMaxScoreEffectCandidate(candidates[i], &wts, &origWts,prevScore,
                                        false, priorMeans, priorStdDevs, error,
                                        NULL, NULL);

      cout << "evaluating gain of candidates took ";
      timer_.printTime(cout, timer_.time()-bsec); cout << endl << endl;


      cout << "finding best candidates..." << endl;
      bsec = timer_.time();
      ClauseOpHashArray* newBeam = new ClauseOpHashArray(beamSize_);
      bool newBestClause = findBestClausesAmongCandidates(candidates, newBeam, 
                                                          bestClauses);
      cout << "finding best candidates took ";
      timer_.printTime(cout, timer_.time()-bsec); cout << endl << endl;

      beam->deleteItemsAndClear();
      delete beam;
      beam = newBeam;

      if (newBestClause)
      {
        iterBestGainUnchanged = 0;
        cout << "found new best clause in beam search iter " << bsiter_ << endl;
      }
      else
        iterBestGainUnchanged++;


        //print the best clauses
      cout << "best clauses found in beam search iter " << bsiter_ << endl;
      cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
      for (int i = 0; i < bestClauses.size(); i++)
      {
        cout << i << "\t";
        bestClauses[i]->printWithoutWtWithStrVar(cout,(*domains_)[0]);
        cout << endl
             << "\tgain = " << bestClauses[i]->getAuxClauseData()->gain
             << ",  op = " 
             << Clause::getOpAsString(bestClauses[i]->getAuxClauseData()->op);
        if (bestClauses[i]->getAuxClauseData()->op != OP_REMOVE)
          cout << ",  wt = " << bestClauses[i]->getWt();
        cout << endl;
      }
      cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl << endl;


      cout << "BEAM SEARCH ITERATION " << bsiter_ << " took ";
      timer_.printTime(cout, timer_.time()-begIterSec); cout << endl << endl;
      cout << "Time elapsed = ";
      timer_.printTime(cout, timer_.time()-startSec_); cout << endl << endl;
    }

    if (beam->empty()) cout << "Beam search ended because beam is empty" 
                            << endl << endl;
    else cout << "Beam search ended because best clause did not change in "
              << iterBestGainUnchanged << " iterations" << endl << endl;
    
    beam->deleteItemsAndClear();
    delete beam;
    bsiter_ = -1;

    if (indexTrans_) indexTrans_->removeClauseIdxToClauseFormulaIdxs(1, 1);

    return numCandidatesConsidered;
  }//beamSearch()


  //This is very similar to regular beam search but it either just
  //adds or just removes predicates from clauses by calling the
  //correct createCandidateClauses method. If addPredicate is true, it
  //will only consider additions. If it is false, it will only
  //consider deletions.
  //returns number of candidates considered
  int pointedBeamSearch(const Array<Clause*>& initClauses, 
			const double& prevScore,
			Array<Clause*>& bestClauses, bool addPredicate, bool justUnaryPreds = false)
  {
    int numConsidered = 0;
    int iterBestGainUnchanged = 0; bsiter_ = -1;
    Array<double> priorMeans, priorStdDevs;
    ClauseOpHashArray* beam = new ClauseOpHashArray;

    //make copies of initclauses so that we don't worry when deleting them
    for (int i = 0; i < initClauses.size(); i++) {
      Clause * copyToAdd = new Clause(*(initClauses[i]));
      if (beam->append(copyToAdd) < 0)
	delete copyToAdd;
    }
    
    assert(beam->size() > 0);

    int numClausesFormulas = getNumClausesFormulas();
    Array<double> wts;
    wts.growToSize(numClausesFormulas + 2); // first slot is unused
    Array<double> origWts(numClausesFormulas); // current clause/formula wts
    if (indexTrans_) indexTrans_->getClauseFormulaWts(origWts);
    else             mln0_->getClauseWts(origWts);

      //set the prior means and std deviations
    setPriorMeansStdDevs(priorMeans, priorStdDevs, 1, NULL);
    if (indexTrans_) indexTrans_->appendClauseIdxToClauseFormulaIdxs(1, 1);

    bool error;
    double begIterSec, bsec;
    Array<Clause*> candidates;
    while (!beam->empty() && iterBestGainUnchanged < bestGainUnchangedLimit_)
    {
      begIterSec = timer_.time();
      bsiter_++; 

      //if (bsiter_ == 1) break; // for testing purposes only
      cout << endl << "BEAM SEARCH ITERATION " << bsiter_ << endl << endl;

      cout << "creating candidate clauses..." << endl;
      candidates.clear();
      bsec = timer_.time();
      
      pointedCreateCandidateClauses(beam, candidates, addPredicate, justUnaryPreds);
      
      numConsidered += candidates.size();

      cout << "num of candidates created = " << candidates.size() << "; took ";
      timer_.printTime(cout, timer_.time()-bsec); cout << endl << endl;

      cout << "evaluating gain of candidates..." << endl << endl;
      bsec = timer_.time();
      for (int i = 0; i < candidates.size(); i++)
        countAndMaxScoreEffectCandidate(candidates[i], &wts, &origWts,prevScore,
                                        false, priorMeans, priorStdDevs, error,
                                        NULL, NULL);
      cout << "evaluating gain of candidates took ";
      timer_.printTime(cout, timer_.time()-bsec); cout << endl << endl;


      cout << "finding best candidates..." << endl;
      bsec = timer_.time();
      ClauseOpHashArray* newBeam = new ClauseOpHashArray(beamSize_);
      bool newBestClause = findBestClausesAmongCandidates(candidates, newBeam, 
                                                          bestClauses);
      cout << "finding best candidates took ";
      timer_.printTime(cout, timer_.time()-bsec); cout << endl << endl;

      beam->deleteItemsAndClear();
      delete beam;
      beam = newBeam;

      if (newBestClause)
      {
        iterBestGainUnchanged = 0;
        cout << "found new best clause in beam search iter " << bsiter_ << endl;
      }
      else
        iterBestGainUnchanged++;


        //print the best clauses
      cout << "best clauses found in beam search iter " << bsiter_ << endl;
      cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
      for (int i = 0; i < bestClauses.size(); i++)
      {
        cout << i << "\t";
        bestClauses[i]->printWithoutWtWithStrVar(cout,(*domains_)[0]);
        cout << endl
             << "\tgain = " << bestClauses[i]->getAuxClauseData()->gain
             << ",  op = " 
             << Clause::getOpAsString(bestClauses[i]->getAuxClauseData()->op);
        if (bestClauses[i]->getAuxClauseData()->op != OP_REMOVE)
          cout << ",  wt = " << bestClauses[i]->getWt();
        cout << endl;
      }
      cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl << endl;


      cout << "BEAM SEARCH ITERATION " << bsiter_ << " took ";
      timer_.printTime(cout, timer_.time()-begIterSec); cout << endl << endl;
      cout << "Time elapsed = ";
      timer_.printTime(cout, timer_.time()-startSec_); cout << endl << endl;
    }

    if (beam->empty()) cout << "Beam search ended because beam is empty" 
                            << endl << endl;
    else cout << "Beam search ended because best clause did not change in "
              << iterBestGainUnchanged << " iterations" << endl << endl;
    
    beam->deleteItemsAndClear();
    delete beam;
    bsiter_ = -1;

    if (indexTrans_) indexTrans_->removeClauseIdxToClauseFormulaIdxs(1, 1);
    return numConsidered;
  }//pointedBeamSearch()

  /////////////// functions dealing with counts and maximizing score ///////////
 private:
  void prepareGibbsAndMWSParams(GibbsParams & gibbsParams, MaxWalksatParams* wsparams) {
    gibbsParams.numChains     = 1;
    gibbsParams.burnMinSteps  = 10;  //expect to be able to burn faster
    gibbsParams.burnMaxSteps  = 400;
    gibbsParams.gamma         = 1-0.05;
    gibbsParams.epsilonFrac   = 0.01;
    gibbsParams.gibbsMinSteps = -1;
    gibbsParams.gibbsMaxSteps = 100;
    gibbsParams.fracConverged = 0.95;
    gibbsParams.walksatType   = MAXWALKSAT;
    gibbsParams.maxSeconds    = -1;
    gibbsParams.samplesPerConvergenceTest = 100;
    
    wsparams->maxSteps = 1000;
    wsparams->tries = 1;
    wsparams->targetWt = INT_MAX;
    wsparams->hard = false; 
  }

 //returns the mrf after diagnostic sampling has been done to
  //it. caller is responsible for deleting the returned pointer
  UTMRF* runDiagnosticInference(GibbsParams gibbsParams, MaxWalksatParams* wsparams,
				Domain * firstInList, const char* nameOfCurrPred, int predId,
				Array<Clause *> trulyInitialClauses) {

    //--Get queries and their truth values
    Array<bool>* truthValues = new Array<bool>;
    GroundPredicateHashArray queries;
    const PredicateTemplate* ptemplate = 
      firstInList->getPredicateTemplate(nameOfCurrPred);
    assert(ptemplate != NULL );

    int varIdCnt = 0;
    Predicate ppred(ptemplate);
    for (int i = 0; i < ptemplate->getNumTerms(); i++)
      ppred.appendTerm(new Term(--varIdCnt, (void*)&ppred, true));

    //create all groundings with this predicate
    Array<VarsTypeId*>* vtiArr = NULL;
    ppred.createVarsTypeIdArr(vtiArr);
    assert(vtiArr->size() > 1);
    ArraysAccessor<int> acc;
    for (int i = 1; i < vtiArr->size(); i++){
      const Array<int>* cons = firstInList->getConstantsByType((*vtiArr)[i]->typeId);
      acc.appendArray(cons);
    }
	
    //temporarily turn off the closed world assumption on this one
    firstInList->getDB()->setClosedWorld(predId, false);
	
    // form all groundings of the predicate
    Array<int> constIds;

    int maxNumberAllowed = 50;
    int totalNumberPresent = acc.numCombinations();

    int numToSkip = 0;
    if (totalNumberPresent > maxNumberAllowed)
      numToSkip = totalNumberPresent / maxNumberAllowed - 1;

    int currentlySkipped = 0;

    while (acc.getNextCombination(constIds)) {
      assert(constIds.size() == vtiArr->size()-1);
      for (int j = 0; j < constIds.size(); j++){
		Array<Term*>& terms = (*vtiArr)[j+1]->vars;
		for (int k = 0; k < terms.size(); k++)
	  	terms[k]->setId(constIds[j]);
      }
    
      GroundPredicate* gndPred = new GroundPredicate(&ppred);

      if (currentlySkipped < numToSkip) {
		if (firstInList->getDB()->getValue(&ppred) == UNKNOWN)
	  		firstInList->getDB()->setValue(&ppred, FALSE);
		delete gndPred;
		currentlySkipped++;
      }
      else {
		currentlySkipped = 0; 
		if (queries.append(gndPred) < 0 ){
		  delete gndPred;
		}
		else{
		  TruthValue tv = firstInList->getDB()->getValue(&ppred);    
		  assert(tv == UNKNOWN || tv == TRUE);
		  truthValues->append((tv==TRUE));
		  
		  //set the value to unknown so we can have some inference
		  //that makes sense. Needs to be restored!
		  firstInList->getDB()->setValue(&ppred, UNKNOWN);
		}   
      }
    }//for each combination of constants
   
    //cout << "Printing queries\n";

    //for (int i = 0; i < queries.size(); i++){
    //  queries[i]->print(cout, firstInList);
    //  cout << endl;
    //}
   
    assert (queries.size() == truthValues->size());

    ppred.deleteVarsTypeIdArr(vtiArr);
    queries.compress();

         
    //--Construct the diagnostic mrf and sample
    UTMRF *theMRF = new UTMRF(&queries, truthValues, firstInList,
			      firstInList->getDB(), mln0_, trulyInitialClauses,
			      false, false);

    cout << "Constructed diagnostic MRF and about to start diagnostic sampling\n";
 
    theMRF->diagnosticGibbsSampling(gibbsParams, wsparams, false, firstInList,
    				    truthValues);
    
    //theMRF->printDiagnosticDataStructures(firstInList);
    cout << "Done with diagnostic gibbs sampling\n";
   
    //--Restore everything back
    for (int q = 0; q < queries.size(); q++){
      Predicate *currPred = queries[q]->createEquivalentPredicate(firstInList);
      if ((*truthValues)[q]) 
		firstInList->getDB()->setValue(currPred, TRUE);
      else
		firstInList->getDB()->setValue(currPred, FALSE);

      assert(firstInList->getDB()->getValue(currPred) != UNKNOWN );

      delete currPred;
    }
     //turn closed world assumption back on 
    firstInList->getDB()->setClosedWorld(predId, true);

    delete truthValues;
    for (int i = 0; i < queries.size(); i++) 
      delete queries[i];
    
    return theMRF;
  }
  
  //L: returns the max gain of any of the clauses
  double computeGainOfClauses(ClauseOpHashArray &newClauses, double & score){
    double maxGain = 0;
    for (int i = 0; i < newClauses.size(); i++){
      Array <double> weights;
      Array <double> priorMeans, priorStdDevs;
      setPriorMeansStdDevs(priorMeans, priorStdDevs, 1, NULL);
      weights.growToSize(getNumClausesFormulas() + 2);
      bool error = false;
      bool resetPriors = false; 

      //cout << "The score before is " << score << endl;

      //this function returns the score, not the gain. We need to get the gain from the aux data of new clauses
      countAndMaxScoreEffectCandidate(newClauses[i], &weights, NULL, score, 
				      resetPriors, priorMeans, priorStdDevs, error,NULL, NULL);

      //cout << "The score after is " << score << endl;

      double currGain = newClauses[i]->getAuxClauseData()->gain;

      //  cout << "The curr score is " << newScore << endl;
      //cout << "The curr gain is " << currGain << endl;
					      
      if (currGain > maxGain || i == 0)
	maxGain = currGain;
    }
    return maxGain;
  }

  //looks for the beamSize_ best clauses in newClauses and adds them
  //to mln. The clauses that are not added are deleted, so there will
  //be holes through this hash array and it should not be used any
  //more. Returns true if the candidates were successfully added to
  //the mln. If something is added, the score is also updated.
  bool findBestAndAddToMLN(ClauseOpHashArray &newClauses, double & score){
    Array<Clause*> candidates;
    for (int i = 0; i < newClauses.size(); i++)
      candidates.append(newClauses[i]);

    Array<Clause*> bestCandidates;

    rquicksort(candidates);    
    
    int numAdded = 0; 
    for (int i = 0; i < candidates.size(); i++){
      //    cout << "numAdded is " << numAdded << endl;
      //cout << "beamSize is " << beamSize_ << endl;
      //cout << "mingain is " << minGain_ << endl;
      //cout << "candidate gain is " << candidates[i]->getAuxClauseData()->gain << endl;
      //cout << "candidate weight is " <<  candidates[i]->getWt() << endl;
      if (numAdded < beamSize_ && 
	  candidates[i]->getAuxClauseData()->gain >= minGain_ &&
	  abs(candidates[i]->getWt()) >= minWt_){
	numAdded++;
	bestCandidates.append(candidates[i]);
      }
      else
	delete candidates[i];
    }

    //cout << "The size of bestCandidates is " << bestCandidates.size() << endl;

    bool ok = true;
    if (bestCandidates.size() > 0)
      ok = appendToAndRemoveFromMLNs(bestCandidates, score);
    
    if (!ok) 
      for (int i = 0; i < bestCandidates.size(); i++)
	delete bestCandidates[i];

    return ok;
  }

  //same as above but takes regular array of clauses as first param.
  bool findBestAndAddToMLN(Array <Clause*> newClauses, double & score){
    Array<Clause*> candidates;
    for (int i = 0; i < newClauses.size(); i++)
      candidates.append(newClauses[i]);

    Array<Clause*> bestCandidates;

    rquicksort(candidates);    
    
    int numAdded = 0; 
    for (int i = 0; i < candidates.size(); i++){
      if (numAdded < beamSize_ && 
	  candidates[i]->getAuxClauseData()->gain >= minGain_ &&
	  abs(candidates[i]->getWt()) >= minWt_){
	numAdded++;
	bestCandidates.append(candidates[i]);
      }
      else
	delete candidates[i];
    }

    //cout << "The size of bestCandidates is " << bestCandidates.size() << endl;

    bool ok = true;
    if (bestCandidates.size() > 0)
      ok = appendToAndRemoveFromMLNs(bestCandidates, score);
    
    if (!ok) 
      for (int i = 0; i < bestCandidates.size(); i++)
	delete bestCandidates[i];

    return ok;
  }


    //size of wts array must be numClauses+1, its first slot is unused
    //if indexTrans_ is used, the sizes of its instance variables must be 
    //correctly set before calling this function
  double maximizeScore(const int& numClausesFormulas, const int& numExtraWts,
                       Array<double>* const & wts, 
                       const Array<double>* const & origWts,
                       const Array<int>* const & removedClauseFormulaIdxs, 
                       int& iter, bool& error, double& elapsedSec)
  {
    if (origWts) 
    { for (int i=1 ; i<=numClausesFormulas; i++) (*wts)[i] = (*origWts)[i-1];}
    else         
    { for (int i=1; i<=numClausesFormulas; i++) (*wts)[i] = 0; }

    for (int i = 1; i <= numExtraWts; i++) (*wts)[numClausesFormulas+i] = 0;

    if (removedClauseFormulaIdxs) 
      for (int i = 0; i < removedClauseFormulaIdxs->size(); i++) 
        (*wts)[ (*removedClauseFormulaIdxs)[i]+1 ] = 0;

    //commented out: this is done before this function
    //if (indexTrans_) 
    //  indexTrans_->appendClauseIdxToClauseFormulaIdxs(numExtraWts, 1);
  
    double* wwts = (double*) wts->getItems();
    double begSec = timer_.time();
    double newScore 
      = lbfgsb_->minimize(numClausesFormulas + numExtraWts, wwts, iter, error);
    newScore = -newScore;
    elapsedSec = timer_.time() - begSec;

    //commented out: this is done after this function
    //if (indexTrans_) 
    //  indexTrans_->removeClauseIdxToClauseFormulaIdxs(numExtraWts, 1);

    return newScore;
  }


    //Compute/remove counts and maximize score when all the candidates are 
    //effected upon MLN. It computes/removes the counts in PseudoLogLikelihood 
    //and then finds the optimal pseudoLogLikelihood and weights, but it  
    //does not add/remove/replace clauses in the MLN.
    //if resetPriors is false, then priorMeans and priorStdDevs must 
    //already contain the mean/std dev of the existing MLN clauses, and be
    //of size #MLN clauses + #candidates
  double countAndMaxScoreEffectAllCandidates(const Array<Clause*>& candidates, 
                                             Array<double>* const & wts,
                                             const Array<double>*const& origWts,
                                             const double& prevScore,
                                             const bool& resetPriors,
                                             Array<double>& priorMeans,
                                             Array<double>& priorStdDevs,
                                             bool& error,
                                            Array<UndoInfo*>* const& uundoInfos,
                         Array<ClauseAndICDArray*>* const & appendedClauseInfos)
  {
    Array<UndoInfo*>* undoInfos = (uundoInfos)? uundoInfos:new Array<UndoInfo*>;
    Array<int> remClauseFormulaIdxs; //for domain 0 only
    Array<int*> idxPtrs; //for clean up
    Array<Clause*> toBeRemovedClauses;
    Array<Clause*> toBeAppendedClauses;
    int numClausesFormulas = getNumClausesFormulas();
      
    for (int i = 0; i < candidates.size(); i++)
    {
      Clause* cand = candidates[i];
      AuxClauseData* acd = cand->getAuxClauseData();
      int op = acd->op; 

        //remove counts for candidate
      if (op == OP_REMOVE || op == OP_REPLACE || op == OP_REPLACE_ADDPRED || 
          op == OP_REPLACE_REMPRED)
      {
        Clause* remClause = (mlns_->size() > 1) ? 
          (Clause*) mln0_->getClause(acd->removedClauseIdx) : NULL;
        Array<int*> idxs; idxs.growToSize(mlns_->size());
        Array<Clause*> remClauses; remClauses.growToSize(mlns_->size());
        for (int d = 0; d < mlns_->size(); d++)
        {
          int remIdx;
          if (d == 0)
          {
            remIdx = acd->removedClauseIdx;
            remClauseFormulaIdxs.append(remIdx);
          }
          else
          {
            if (remClause->containsConstants())
              remClause->translateConstants((*domains_)[d-1], (*domains_)[d]);
            remIdx = (*mlns_)[d]->findClauseIdx(remClause);
          }
          idxs[d] = (*mlns_)[d]->getMLNClauseInfoIndexPtr(remIdx);
          remClauses[d] = (Clause*) (*mlns_)[d]->getClause(remIdx);
        }
        if (mlns_->size() > 1 && remClause->containsConstants())
          remClause->translateConstants((*domains_)[mlns_->size()-1], 
                                        (*domains_)[0]);
        pllRemoveCountsForClause(remClauses, idxs, undoInfos);

        if (op == OP_REMOVE) toBeRemovedClauses.append(cand);
      }

        //compute counts for candidate
      if (op == OP_ADD || op == OP_REPLACE || op == OP_REPLACE_ADDPRED || 
          op == OP_REPLACE_REMPRED)
      {
        Array<int*> tmpIdxs; tmpIdxs.growToSize(mlns_->size());
        for (int d = 0; d < mlns_->size(); d++)
        {
          int* tmpClauseIdxInMLN = new int( (*mlns_)[d]->getNumClauses() + 
                                            toBeAppendedClauses.size() );
          tmpIdxs[d] = tmpClauseIdxInMLN;
          idxPtrs.append(tmpClauseIdxInMLN);
        }

        Array<UndoInfo*>* tmpInfos = (appendedClauseInfos)?  
                                     new Array<UndoInfo*> : undoInfos;
        pllComputeCountsForClause(cand, tmpIdxs, tmpInfos); 
        toBeAppendedClauses.append(cand);
        
        if (appendedClauseInfos) 
        {
          undoInfos->append(*tmpInfos);
          ClauseAndICDArray* ca = new ClauseAndICDArray;
          ca->clause = cand;
          appendedClauseInfos->append(ca);          
          for (int j = 0; j < tmpInfos->size(); j++)
          {
            ca->icdArray.append(IndexCountDomainIdx());
            ca->icdArray.lastItem().iac=(*tmpInfos)[j]->affectedArr->lastItem();
            ca->icdArray.lastItem().domainIdx = (*tmpInfos)[j]->domainIdx;
          }
          delete tmpInfos;
        }
      }      
    }

    //find optimal weights and score

      // find the clause/formula indexes for indexes in remClauseFormulaIdxs 
    if (indexTrans_) 
      indexTrans_->getClauseFormulaIndexes(remClauseFormulaIdxs, 0);
    
    Array<double> removedValues;
    if (resetPriors) 
    {
      setPriorMeansStdDevs(priorMeans, priorStdDevs, toBeAppendedClauses.size(),
                           &remClauseFormulaIdxs);
      if (indexTrans_) indexTrans_->appendClauseIdxToClauseFormulaIdxs(
                                                 toBeAppendedClauses.size(), 1);
    }
    else
    {
      assert(priorMeans.size() == numClausesFormulas + candidates.size());
      assert(priorStdDevs.size() == numClausesFormulas + candidates.size());
      if (indexTrans_) 
        assert(indexTrans_->checkCIdxWtsGradsSize(candidates.size()));

      setRemoveAppendedPriorMeans(numClausesFormulas, priorMeans, 
                                  remClauseFormulaIdxs, 
                                  toBeAppendedClauses.size(), removedValues);
    }

    if (hasPrior_) 
      pll_->setMeansStdDevs(priorMeans.size(), priorMeans.getItems(), 
                            priorStdDevs.getItems());
    else
      pll_->setMeansStdDevs(-1, NULL, NULL);

    wts->growToSize(numClausesFormulas + toBeAppendedClauses.size() + 1);

    int iter; double elapsedSec; 
    double newScore = maximizeScore(numClausesFormulas, 
                                    toBeAppendedClauses.size(), 
                                    wts, origWts, &remClauseFormulaIdxs,
                                    iter, error, elapsedSec);

      //set the weights of the candidates
    for (int i = 0; i < toBeAppendedClauses.size(); i++)
      toBeAppendedClauses[i]->setWt( (*wts)[numClausesFormulas+i+1] );
      //set to large weight so that they are filtered out because of small wt
    for (int i = 0; i < toBeRemovedClauses.size(); i++)
      toBeRemovedClauses[i]->setWt(111111111); 
    
    Array<double> penalties;
    for (int i = 0; i < candidates.size(); i++)
      penalties.append(getPenalty(candidates[i]));
                     
    if (error) { newScore=prevScore; cout<<"LBFGSB failed to find wts"<<endl; }
   printNewScore(candidates, (*domains_)[0], iter, elapsedSec, 
		  newScore, newScore-prevScore, penalties);

    for (int i = 0; i < candidates.size(); i++)
      candidates[i]->getAuxClauseData()->gain = newScore-prevScore-penalties[i];

    if (uundoInfos == NULL) 
    { 
      pll_->undoAppendRemoveCounts(undoInfos); 
      delete undoInfos;
    }

    if (resetPriors) 
    {
      if (indexTrans_) indexTrans_->removeClauseIdxToClauseFormulaIdxs(
                                                 toBeAppendedClauses.size(), 1);
    }
    else
    {   //restore the prior means that were set to zero
      for (int i = 0; i < remClauseFormulaIdxs.size(); i++)
        priorMeans[remClauseFormulaIdxs[i]] = removedValues[i];
    }

    idxPtrs.deleteItemsAndClear();

    return newScore;
  } //countAndMaxScoreEffectAllCandidates()


  double countAndMaxScoreEffectCandidate(Clause* const & candidate, 
                                         Array<double>* const & wts,
                                         const Array<double>* const & origWts, 
                                         const double& prevScore,
                                         const bool& resetPriors,
                                         Array<double>& priorMeans,
                                         Array<double>& priorStdDevs,
                                         bool& error,
                                         Array<UndoInfo*>* const& undoInfos,
                         Array<ClauseAndICDArray*>* const & appendedClauseInfos)
  {
    Array<Clause*> candidates; candidates.append(candidate);
    return countAndMaxScoreEffectAllCandidates(candidates, wts, origWts, 
                                               prevScore, resetPriors,
                                               priorMeans, priorStdDevs, error,
                                               undoInfos, appendedClauseInfos);
  }


    //returns true if the candidates are all effected on MLN resulting in a
    //better score
  bool appendToAndRemoveFromMLNs(const Array<Clause*>& candidates,double& score,
                                const bool& makeChangeForEqualScore=false)
  {
    Array<double> wts;
    Array<UndoInfo*> undoInfos;
    Array<ClauseAndICDArray*> appendedClauseInfos;
    Array<double> priorMeans, priorStdDevs;
    bool error;

    double newScore
      = countAndMaxScoreEffectAllCandidates(candidates, &wts, NULL, score,
                                            true, priorMeans,priorStdDevs,error,
                                            &undoInfos, &appendedClauseInfos);

    bool improve = (makeChangeForEqualScore) ? (newScore >= score) 
                                             : (newScore > score);

    if (!error && improve)
    {
      score = newScore;

        //set MLN clauses/formulas to new weights
        //weights of appended clauses are set when they are added to the MLN 
      updateWts(wts, NULL, NULL);

      int numClausesFormulas = getNumClausesFormulas();

      for (int i = 0; i < appendedClauseInfos.size(); i++)
        appendedClauseInfos[i]->clause->setWt(wts[numClausesFormulas+i+1]);

        //effect candidates on MLN
      int appClauseIdx = 0;
      for (int i = 0; i < candidates.size(); i++)
      {
        Clause* cand = candidates[i];
        AuxClauseData* acd = cand->getAuxClauseData();
        int op = acd->op;

          // NOTE: cand was not canonicalized so as to easily compare it to the
          //       mln clause it is to replace. Canonicalize it now before 
          //       appending to MLN.
        if (op == OP_REPLACE) cand->canonicalize();

        if (op == OP_REMOVE || op == OP_REPLACE || op == OP_REPLACE_ADDPRED || 
            op == OP_REPLACE_REMPRED)
        {
          Clause* remClause = (Clause*) mln0_->getClause(acd->removedClauseIdx);

          for (int d = 0; d < mlns_->size(); d++)
          {
            if (d == 0)
            {
              Clause* r = removeClauseFromMLN(acd->removedClauseIdx, d);
              cout << "Modified MLN: Removed clause from MLN: "; 
              r->printWithoutWtWithStrVar(cout,(*domains_)[0]); cout << endl;
              if (op == OP_REMOVE && cand != r) delete cand;
              //delete r; //this is remClause which has to be used below
            }
            else
            {
              if (remClause->containsConstants())
                remClause->translateConstants((*domains_)[d-1], (*domains_)[d]);
              int remIdx = (*mlns_)[d]->findClauseIdx(remClause);
              delete removeClauseFromMLN(remIdx, d);
            }
          }
          delete remClause;
        }
        
        if (op == OP_ADD || op == OP_REPLACE || op == OP_REPLACE_ADDPRED || 
            op == OP_REPLACE_REMPRED)
        {
          Array<int*> idxPtrs; idxPtrs.growToSize(mlns_->size());
          for (int d = 0; d < mlns_->size(); d++)
          {
            Clause* c = cand;
            if (d > 0)
            {
              if (cand->containsConstants()) 
                cand->translateConstants((*domains_)[d-1], (*domains_)[d]);
              c = new Clause(*cand);
            }
            int idx = appendClauseToMLN(c, d);
            idxPtrs[d] = (*mlns_)[d]->getMLNClauseInfoIndexPtr(idx);
          }

          Array<IndexCountDomainIdx>& icds 
            = appendedClauseInfos[appClauseIdx++]->icdArray;
          for (int j = 0; j < icds.size(); j++)
            icds[j].iac->index = idxPtrs[ icds[j].domainIdx ];
            
          cout << "Modified MLN: Appended clause to MLN: "; 
          cand->printWithoutWtWithStrVar(cout,(*domains_)[0]); 
          cout << endl;
        }
      }

      assert(pll_->checkNoRepeatedIndex());
      assert(appClauseIdx == appendedClauseInfos.size());

      undoInfos.deleteItemsAndClear();

        //MLNs has changed, the index translation must be recreated
      if (indexTrans_) indexTrans_->createClauseIdxToClauseFormulaIdxsMap();
    }
    else
    {
      cout << "undoing candidates because score did not improve..."<<endl<<endl;
      pll_->undoAppendRemoveCounts(&undoInfos);
    }
    
    appendedClauseInfos.deleteItemsAndClear();
    return improve;
  }//appendToAndRemoveFromMLNs()


    //returns true if candidate is effected on MLN resulting in a better score
  bool appendToAndRemoveFromMLNs(Clause* const & candidate, double& score,
                                 const bool& makeChangeForEqualScore=false)
  {
    Array<Clause*> candidates; candidates.append(candidate);
    return appendToAndRemoveFromMLNs(candidates, score,makeChangeForEqualScore);
  }


    //Clause c enters & exits  function with its AuxClauseData's cache == NULL
  void pllComputeCountsForClause(Clause* const & c,
                                 const Array<int*>& clauseIdxInMLNs,
                                 Array<UndoInfo*>* const & undoInfos)
  {
    assert(c->getAuxClauseData()->cache == NULL);
    assert(clauseIdxInMLNs.size() == domains_->size());
    double begSec = timer_.time();
    
    if (cacheClauses_)    
    {
      int i;
      if ((i = cachedClauses_->find(c)) >= 0) // if clause and counts are cached
      {
        //cout << "using cached counts ";
        //cout << "for "; c->printWithoutWtWithStrVar(cout, (*domains_)[0]); 
        //cout << endl;
        pll_->insertCounts(clauseIdxInMLNs, undoInfos, 
                           (*cachedClauses_)[i]->getAuxClauseData()->cache);
        cout << "using cached counts took ";
        timer_.printTime(cout, timer_.time()-begSec); cout << endl;
        return;
      }
      else
      {
        assert(c->getAuxClauseData()->cache == NULL);
        if (cacheSizeMB_ <  maxCacheSizeMB_)
          c->newCache(domains_->size(), (*domains_)[0]->getNumPredicates());
        else
        {
          static bool printCacheFull = true;
          if (printCacheFull)
          {
            cout << "Cache is full, approximate size = " << cacheSizeMB_ 
                 << " MB" << endl;
            printCacheFull = false;
          }
        }
        //cout << "cache size (MB) = " << cacheSizeMB_ << endl;
      }
    }
    
    //we do not have the counts

      //if we don't need to translate constant ids from one domain to another
    if (!c->containsConstants())
    {
      for (int i = 0; i < domains_->size(); i++)
      {
        int* clauseIdxInMLN = clauseIdxInMLNs[i];
        pll_->computeCountsForNewAppendedClause(c, clauseIdxInMLN, i,
                                                undoInfos, sampleClauses_, 
                                                c->getAuxClauseData()->cache);
      }
    }
    else
    {  //we need to translate constant ids from one domain to another
      int i; 
      for (i = 0; i < domains_->size(); i++)
      {
        if (i > 0) c->translateConstants((*domains_)[i-1], (*domains_)[i]);
        int* clauseIdxInMLN = clauseIdxInMLNs[i];
        pll_->computeCountsForNewAppendedClause(c, clauseIdxInMLN, i,
                                                undoInfos, sampleClauses_, 
                                                c->getAuxClauseData()->cache);
      }
      if (i > 1) c->translateConstants((*domains_)[i-1], (*domains_)[0]);
    }


    if (c->getAuxClauseData()->cache)
    {
      if (cacheSizeMB_ < maxCacheSizeMB_)
      {
        cacheSizeMB_ += c->sizeMB();
        Array<Array<Array<CacheCount*>*>*>* cache =c->getAuxClauseData()->cache;
        c->getAuxClauseData()->cache = NULL;
        Clause* copyClause = new Clause(*c);
        copyClause->getAuxClauseData()->cache = cache;
        copyClause->compress();
        cachedClauses_->append(copyClause);
      }
      else
      {
        c->getAuxClauseData()->deleteCache();
        c->getAuxClauseData()->cache = NULL; 
      }
    }

    // cout << "Computing counts took ";
    //timer_.printTime(cout, timer_.time()-begSec); 
    //cout << " for " << endl ;
    //cout << "\t"; c->printWithoutWtWithStrVar(cout, (*domains_)[0]); 
    //cout << endl;
  }


  void pllRemoveCountsForClause(const Array<Clause*>& remClauses, 
                                const Array<int*>& clauseIdxInMLNs,
                                Array<UndoInfo*>* const & undoInfos)
  { 
    assert(clauseIdxInMLNs.size() == domains_->size());
    //    double begSec = timer_.time();        
    for (int i = 0; i < domains_->size(); i++)
      pll_->removeCountsForClause(remClauses[i],clauseIdxInMLNs[i],i,undoInfos);
    //cout << "Removing counts took ";
    //timer_.printTime(cout, timer_.time()-begSec); 
    //cout << endl;
  }



  void pllComputeCountsForInitialMLN()
  {
    for (int i = 0; i < mlns_->size(); i++)
    {
      cout << "computing counts for clauses in domain " << i << "..." << endl;
      MLN* mln = (*mlns_)[i];
      for (int j = 0; j < mln->getNumClauses(); j++)
      {
        Clause* c = (Clause*)mln->getClause(j);
        cout << "Clause " << j << ": ";
        c->printWithoutWtWithStrVar(cout, (*domains_)[i]); cout << endl;
        int* clauseIdxInMLN = mln->getMLNClauseInfoIndexPtr(j);
        pll_->computeCountsForNewAppendedClause(c, clauseIdxInMLN, i,
                                            NULL, sampleClauses_, NULL);
      }
    }
  }


  //////////////////// adding/removing/replacing MLN clauses /////////////////
 private:
  void addUnitClausesToMLNs()
  {
    Array<Predicate*> nonEvidPreds;
    for (int i = 0; i < preds_->size(); i++)
      if ((*areNonEvidPreds_)[(*preds_)[i]->getId()])
        nonEvidPreds.append((*preds_)[i]);

    Array<Clause*> unitClauses;
    bool allowEqualPreds = true;
    ClauseFactory::createUnitClauses(unitClauses,nonEvidPreds,allowEqualPreds);

    for (int i = 0; i < unitClauses.size(); i++)
    {
      if (mln0_->containsClause(unitClauses[i]))
      { delete unitClauses[i]; continue; }
      ostringstream oss; int idx; 
      unitClauses[i]->printWithoutWtWithStrVar(oss, (*domains_)[0]);
      
      for (int j = 0; j < mlns_->size(); j++)
      {
        Clause* c = (j == 0) ? unitClauses[i] : new Clause(*unitClauses[i]);
        (*mlns_)[j]->appendClause(oss.str(), false, c, priorMean_, false, idx);
        ((MLNClauseInfo*)(*mlns_)[j]->getMLNClauseInfo(idx))->priorMean 
          = priorMean_;
      }
    }
  }


  void appendUnitClausesWithDiffCombOfVar(double& score)
  {
    bool allowEqualPreds = false;
    for (int i = 0; i < preds_->size(); i++)
    {
      if (!(*areNonEvidPreds_)[(*preds_)[i]->getId()]) continue;

      Clause* origClause = ClauseFactory::createUnitClause((*preds_)[i], 
                                                           allowEqualPreds);
      if (origClause == NULL) continue;
      assert(origClause->getAuxClauseData() == NULL);
      origClause->setAuxClauseData(new AuxClauseData);
      origClause->trackConstants();

      ClauseOpHashArray newUnitClauses;
      clauseFactory_->createUnitClausesWithDiffCombOfVar((*preds_)[i], OP_ADD,
                                                         -1, newUnitClauses);
      for (int j = 0; j < newUnitClauses.size(); j++)
      {
        Clause* newClause = newUnitClauses[j];

        if (origClause->same(newClause) || 
            !clauseFactory_->validClause(newClause) ||
            mln0_->containsClause(newClause))
        {
          newUnitClauses.removeItemFastDisorder(j);
          delete newClause; 
          j--;
          continue;
        }
          //score is updated
        if (!appendToAndRemoveFromMLNs(newClause, score)) delete newClause;
      }
      delete origClause;
    } // for each predicate
  }


  void flipMLNClauses(double& score)
  {
    Array<Clause*> mlnClauses;
    for (int i = 0; i < mln0_->getNumClauses(); i++)
    {
        //do not flip unit clauses or those derived from existential formulas
      if (mln0_->getClause(i)->getNumPredicates()==1 || !isModifiableClause(i))
        continue;
      mlnClauses.append((Clause*)mln0_->getClause(i));
    }

    for (int i = 0; i < mlnClauses.size(); i++)
    {
      Clause* origClause = mlnClauses[i];
      int origIdx = mln0_->findClauseIdx(origClause);

        // NOTE: new clauses are not canonicalized so that they can be easily 
        //       compared with the original to determine its penalty.
      bool canonicalizeNewClauses = false;
      ClauseOpHashArray newClauses;
      clauseFactory_->flipSensesInClause(origClause, OP_REPLACE, origIdx,
                                         newClauses, canonicalizeNewClauses);

      Array<double> priorMeans, priorStdDevs;
      setPriorMeansStdDevs(priorMeans, priorStdDevs, 1, NULL);
      Array<double> wts;
      wts.growToSize(getNumClausesFormulas()+2);
      Clause* bestClause = NULL;
      double bestScore = score;
      bool error;

      for (int j = 0; j < newClauses.size(); j++) // for each 'flipped' clause
      {
        Clause* newClause = newClauses[j];
        if (origClause->same(newClause) || newClause->hasRedundantPredicates() 
            || mln0_->containsClause(newClause)) { delete newClause; continue; }

        double newScore 
          = countAndMaxScoreEffectCandidate(newClause, &wts, NULL, score, false,
                                            priorMeans, priorStdDevs, error, 
                                            NULL, NULL);
        if (newScore > bestScore)
        {
          bestScore = newScore;
          if (bestClause) delete bestClause;
          bestClause = newClause;          
        }
        else
          delete newClause;
      }

      if (bestClause && !appendToAndRemoveFromMLNs(bestClause, score)) 
        delete bestClause;
    }// for each MLN clause
  }//flipMLNClauses()


    //NOTE: formulas with existentially quantified variables, or variables
    //      with mutually exclusive and exhaustive values are not pruned.
  void pruneMLN(double& score)
  {
    Array<Clause*> mlnClauses;
    for (int i = 0; i < mln0_->getNumClauses(); i++)
    {
        //do not prune unit clauses or clauses derived from existential formulas
      if (mln0_->getClause(i)->getNumPredicates() == 1 || 
          !isModifiableClause(i)) continue;
      mlnClauses.append((Clause*)mln0_->getClause(i));
    }
    
    for (int i = 0; i < mlnClauses.size(); i++)
    {
      Clause* origClause = mlnClauses[i];
      int origIdx = mln0_->findClauseIdx(origClause);
      Clause* copy = new Clause(*origClause);
      copy->setAuxClauseData(new AuxClauseData(0, OP_REMOVE, origIdx));
      copy->trackConstants();
      if (!appendToAndRemoveFromMLNs(copy, score, true)) delete copy;
    }
  }


  int appendClauseToMLN(Clause* const c, const int& domainIdx)
  {
    ostringstream oss; int idx;
    c->printWithoutWtWithStrVar(oss, (*domains_)[domainIdx]);
    MLN* mln = (*mlns_)[domainIdx];
    bool ok = mln->appendClause(oss.str(), false, c, c->getWt(), false, idx);
    if (!ok) { cout << "ERROR: failed to insert " << oss.str() <<" into MLN"
                    << endl; exit(-1); }
    ((MLNClauseInfo*)mln->getMLNClauseInfo(idx))->priorMean = priorMean_;
    return idx;
  }


  Clause* removeClauseFromMLN(const int& remClauseIdx, const int& domainIdx)
  {
    Clause* remClause = (*mlns_)[domainIdx]->removeClause(remClauseIdx);
    if (remClause == NULL)
    { 
      cout << "ERROR: failed to remove "; 
      remClause->printWithoutWtWithStrVar(cout, (*domains_)[0]);
      cout << " from MLN" << endl;
      exit(-1);
    }
    return remClause;
  }

  
  ////////////////////// functions for clause creation ////////////////////////
 private:

  void addPredicateToClause(Clause* const & beamClause, const int& op, 
                            const int& removeClauseIdx,
                            ClauseOpHashArray& newClauses, bool justUnaryPreds = false)
  {
    Array<Predicate *> * predsToUse;
    if (justUnaryPreds) {
      cout << "&&%%&& Doing just unary predicates\n";
      predsToUse = new Array<Predicate *>; //needs to be deleted
      for (int i = 0; i < preds_->size(); i++) 
	if ((*preds_)[i]->getNumTerms() == 1)
	  predsToUse->append((*preds_)[i]);
    }
    else 
      predsToUse = preds_;

      // create new clauses by adding predicates to beamClause
    clauseFactory_->addPredicateToClause(*predsToUse,beamClause,op,removeClauseIdx,
                                         true,newClauses,false);
      // if this is a unit clause, flip its sense and create new clauses by
      // adding preds to it
    if (beamClause->getNumPredicates() == 1)
    {
      beamClause->getPredicate(0)->invertSense();        
      clauseFactory_->addPredicateToClause(*predsToUse, beamClause, op,
                                           removeClauseIdx, true, newClauses,
                                           false);
      beamClause->getPredicate(0)->invertSense();
    }    


    if (justUnaryPreds)
      delete predsToUse;
  }


    // create new clauses by removing a predicate from beamClause
  void removePredicateFromClause(Clause* const & beamClause, const int& op,
                                 const int& removeClauseIdx,
                                 ClauseOpHashArray& newClauses)
  {
    if (beamClause->getNumPredicates() > 2){
      cout << "Going into clause factory's version of removePred\n";
      clauseFactory_->removePredicateFromClause(beamClause, op, removeClauseIdx,
                                                newClauses);
    }
  }


  //L : method is similar to above version. Used for smart-learning
  void removeParticularPredicateFromClause(const int& predIdx,
					   const Clause * const & clause,
					   const int & op, const int & removeClauseIdx,
					   ClauseOpHashArray & newClauses){
    if (clause->getNumPredicates() > 2){
      clauseFactory_->removePredicateFromClause(predIdx, clause, op, removeClauseIdx, newClauses);
    }
}


  void addNewClauseToCandidates(Clause* const & newClause,
                                Array<Clause*>& candidates, 
                                Array<Clause*>* const & thrownOut)
  {
      // remove any new clauses that are already in MLN
    if (thrownOut && mln0_->containsClause(newClause)) 
    { thrownOut->append(newClause); return;}
    newClause->setWt(0);
    candidates.append(newClause);
  }

  //L: This goes with the pointed beam search. Either only adds
  //predicates from clauses or only removes them.
  void pointedCreateCandidateClauses(const ClauseOpHashArray* const & beam, 
				     Array<Clause*>& candidates, 
				     bool addPredicate, bool justUnaryPreds = false){
    cout << "Now in pointedCreateCandidateClauses method\n";
    cout << "The value of justUnaryPreds is " << justUnaryPreds << endl;
      // new clauses that are thrown out because they are in MLN
    Array<Clause*> thrownOutClauses; 
    ClauseOpHashArray newClauses;
    for (int i = 0; i < beam->size(); i++)
    {
      Clause* beamClause = (*beam)[i];
      AuxClauseData* beamacd = beamClause->getAuxClauseData();      
      int op = beamacd->op;
      int newClausesBegIdx = newClauses.size();

      assert (op == OP_ADD || op == OP_NONE);

      if (op == OP_ADD)
      {
        int remIdx = beamacd->removedClauseIdx;

	if (addPredicate)
	  addPredicateToClause(beamClause, op, remIdx, newClauses, justUnaryPreds);
	else
	  removePredicateFromClause(beamClause, op, remIdx, newClauses);
       
	for (int j = newClausesBegIdx; j < newClauses.size(); j++)
          addNewClauseToCandidates(newClauses[j], candidates,&thrownOutClauses);
      }
      else
      if (op == OP_NONE)
      {
        //int idx = mln0_->findClauseIdx(beamClause);
	if (addPredicate) {
	  addPredicateToClause(beamClause, OP_ADD, -1, newClauses, justUnaryPreds);
	  for (int j = newClausesBegIdx; j < newClauses.size(); j++){
	    addNewClauseToCandidates(newClauses[j], candidates,
				     &thrownOutClauses);
	  }
	
          newClausesBegIdx = newClauses.size();          
          if (!addPredicate) {
	    removePredicateFromClause(beamClause, OP_ADD, -1, newClauses);
	    for (int j = newClausesBegIdx; j < newClauses.size(); j++){
	      addNewClauseToCandidates(newClauses[j], candidates,
				       &thrownOutClauses);
	    }
          }
        }
      }
    } // for each item in beam
 }


  void createCandidateClauses(const ClauseOpHashArray* const & beam, 
                              Array<Clause*>& candidates,
                              const Array<Clause*>* const & initMLNClauses, bool justUnaryPreds = false)
  {
      // new clauses that are thrown out because they are in MLN
    Array<Clause*> thrownOutClauses; 
    ClauseOpHashArray newClauses;
    for (int i = 0; i < beam->size(); i++)
    {
      Clause* beamClause = (*beam)[i];
      AuxClauseData* beamacd = beamClause->getAuxClauseData();      
      int op = beamacd->op;
      int newClausesBegIdx = newClauses.size();

      if (op == OP_ADD)
      {
        int remIdx = beamacd->removedClauseIdx;
        addPredicateToClause(beamClause, op, remIdx, newClauses, justUnaryPreds);
        removePredicateFromClause(beamClause, op, remIdx, newClauses);
        for (int j = newClausesBegIdx; j < newClauses.size(); j++)
          addNewClauseToCandidates(newClauses[j], candidates,&thrownOutClauses);
      }
      else
      if (op == OP_REPLACE_ADDPRED)
      {
        addPredicateToClause(beamClause, op, beamacd->removedClauseIdx,
                             newClauses, justUnaryPreds);
        for (int j = newClausesBegIdx; j < newClauses.size(); j++)
          addNewClauseToCandidates(newClauses[j], candidates,&thrownOutClauses);
      }
      else
      if (op == OP_REPLACE_REMPRED)
      {
        removePredicateFromClause(beamClause, op, beamacd->removedClauseIdx,
                                  newClauses);
        for (int j = newClausesBegIdx; j < newClauses.size(); j++)
          addNewClauseToCandidates(newClauses[j], candidates,&thrownOutClauses);
      }
      else
      if (op == OP_NONE)
      {
        int idx = mln0_->findClauseIdx(beamClause);
        bool beamClauseInMLN = (idx >= 0);
        bool removeBeamClause = (beamClauseInMLN && 
                                 beamClause->getNumPredicates() > 1);
        bool isModClause = (!beamClauseInMLN || isModifiableClause(idx));
        
        if (isModClause)
        {
          addPredicateToClause(beamClause, OP_ADD, -1, newClauses, justUnaryPreds);
          for (int j = newClausesBegIdx; j < newClauses.size(); j++)
          {
            addNewClauseToCandidates(newClauses[j], candidates,
                                     &thrownOutClauses);
            if (removeBeamClause)
            {
              Clause* c = new Clause(*(newClauses[j]));
              c->getAuxClauseData()->op = OP_REPLACE_ADDPRED;
              c->getAuxClauseData()->removedClauseIdx = idx;
              addNewClauseToCandidates(c, candidates, &thrownOutClauses);
            }
          }

          newClausesBegIdx = newClauses.size();          
          removePredicateFromClause(beamClause, OP_ADD, -1, newClauses);
          for (int j = newClausesBegIdx; j < newClauses.size(); j++)
          {
            addNewClauseToCandidates(newClauses[j], candidates,
                                     &thrownOutClauses);
            if (removeBeamClause)
            {
              Clause* c = new Clause(*(newClauses[j]));
              c->getAuxClauseData()->op = OP_REPLACE_REMPRED;
              c->getAuxClauseData()->removedClauseIdx = idx;
              addNewClauseToCandidates(c, candidates, &thrownOutClauses);
            }
          }

          if (removeBeamClause)
          {
            Clause* c = new Clause(*beamClause);
            c->getAuxClauseData()->op = OP_REMOVE;
            c->getAuxClauseData()->removedClauseIdx = idx;
            addNewClauseToCandidates(c, candidates, NULL);
          }
        }
      }
      else
        assert(op == OP_REMOVE || op == OP_REPLACE);
    } // for each item in beam


    if (initMLNClauses)
    {
        // add the MLN clauses to the candidates in the first step of beam 
        // search when we start from an empty MLN
      int newClausesBegIdx = newClauses.size();
      for (int i = 0; i < initMLNClauses->size(); i++)
      {
        Clause* c = new Clause(*((*initMLNClauses)[i]));
        if (newClauses.append(c) < 0) delete c;
      }
      for (int i = newClausesBegIdx; i < newClauses.size(); i++)
        addNewClauseToCandidates(newClauses[i], candidates, &thrownOutClauses);
    }

    for (int i = 0; i < thrownOutClauses.size(); i++) 
      delete thrownOutClauses[i];
  }


  //////////////////////// helper functions ////////////////////////////////

 private:
  void useTightParams()
  {
    sampleClauses_ = false;
    pll_->setSampleGndPreds(false);
    lbfgsb_->setMaxIter(lbMaxIter_);
    lbfgsb_->setFtol(lbConvThresh_);
    cacheClauses_ = false;
  }


  void useLooseParams()
  {
    sampleClauses_ = origSampleClauses_;
    pll_->setSampleGndPreds(sampleGndPreds_);
    if (looseMaxIter_ >= 0) lbfgsb_->setMaxIter(looseMaxIter_);
    if (looseConvThresh_ >= 0) lbfgsb_->setFtol(looseConvThresh_);
    cacheClauses_ = origCacheClauses_;
  }

  
    //i is the clause's index in mln_[0]
  bool isModifiableClause(const int& i) const
  { return (!mln0_->isExistClause(i) && !mln0_->isExistUniqueClause(i)); }


  bool isNonModifiableFormula(const FormulaAndClauses* const & fnc) const
  { return (fnc->hasExist || fnc->isExistUnique); }


  void printIterAndTimeElapsed(const double& begSec)
  {
    cout << "Iteration " << iter_ << " took ";
    timer_.printTime(cout, timer_.time()-begSec); cout << endl << endl;
    cout << "Time elapsed = ";
    timer_.printTime(cout, timer_.time()-startSec_); cout << endl << endl;
  }


  void removeClausesFromMLNs()
  {
    for (int i = 0; i < mlns_->size(); i++)
      (*mlns_)[i]->removeAllClauses(NULL);
  }


  void reEvaluateCandidates(Array<Clause*>& candidates,
                            const double& prevScore)
  {
    int numClausesFormulas = getNumClausesFormulas();
    Array<double> wts; 
    wts.growToSize(numClausesFormulas + 2);//slot 0 is unused

    Array<double> priorMeans, priorStdDevs;
    setPriorMeansStdDevs(priorMeans, priorStdDevs, 1, NULL);
    if (indexTrans_) indexTrans_->appendClauseIdxToClauseFormulaIdxs(1, 1);

    bool error;
    for (int i = 0; i < candidates.size(); i++)
    {
      candidates[i]->getAuxClauseData()->gain = 0;
      countAndMaxScoreEffectCandidate(candidates[i], &wts, NULL, prevScore,
                                      false, priorMeans, priorStdDevs, error,
                                      NULL, NULL);
    }

    if (indexTrans_) indexTrans_->removeClauseIdxToClauseFormulaIdxs(1, 1);

    Array<Clause*> tmpCand(candidates);
    rquicksort(tmpCand);
    candidates.clear();
    for (int i = 0; i < tmpCand.size(); i++)
    {
      if (tmpCand[i]->getAuxClauseData()->gain > minGain_ && 
          fabs(tmpCand[i]->getWt()) >= minWt_)
        candidates.append(tmpCand[i]);
      else
        delete tmpCand[i];
    }

    cout << "reevaluated top " << candidates.size() 
         << " candidates with tight params:" << endl;
    cout << "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@" << endl;
    for (int i = 0; i < candidates.size(); i++)
    {
      cout << i << "\t";
      candidates[i]->printWithoutWtWithStrVar(cout,(*domains_)[0]);
      cout << endl 
           << "\tgain = " << candidates[i]->getAuxClauseData()->gain 
           << ",  wt = " << candidates[i]->getWt()
           << ",  op = "
           << Clause::getOpAsString(candidates[i]->getAuxClauseData()->op) 
           << endl;
    }
    cout << "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@" << endl << endl;;
  }


    //Copy the best beamSize_ candidates into beam, and place the best 
    //numEstBestClauses_ clauses among those in candidates and bestClauses into
    //bestClauses. Returns true if the best clause changed.
  bool findBestClausesAmongCandidates(Array<Clause*>& candidates,
                                      ClauseOpHashArray* const & beam,
                                      Array<Clause*>& bestClauses)
  {
    assert(beam->size() == 0);

    // fill the beam with best beamSize_ candidates

      //sort candidates in order of decreasing gain
    rquicksort(candidates); 

    for (int i = 0; i < candidates.size(); i++)
    {
      if (beam->size() >= beamSize_) break;
      double candGain = candidates[i]->getAuxClauseData()->gain;
      double candAbsWt = fabs(candidates[i]->getWt());
      if (candGain > minGain_ && candAbsWt >= minWt_)
      {
        int a = beam->append(new Clause(*(candidates[i])));
        assert(a >= 0); a = 0; //avoid compilation warning
      }
    }
    
    if (beam->size() == 0)
    {
      for (int i = 0; i < candidates.size(); i++) delete candidates[i];
      return false;
    }


    Clause* prevBestClause = (bestClauses.size() > 0) ? 
                             new Clause(*(bestClauses[0])) : NULL;

    // pick the best numEstBestClauses_ clauses

    ClauseOpSet cset;
    ClauseOpSet::iterator it;
    for (int i = 0; i < candidates.size(); i++) cset.insert(candidates[i]);

    for (int i = 0; i < bestClauses.size(); i++) 
    {
      if ((it=cset.find(bestClauses[i])) == cset.end())
      {
        candidates.append(bestClauses[i]);
        cset.insert(bestClauses[i]);
      }
      else
      {
        assert((*it)->getAuxClauseData()->gain == 
               bestClauses[i]->getAuxClauseData()->gain);
        delete bestClauses[i];
      }
    }
    
    rquicksort(candidates);

    bestClauses.clear();
    for (int i = 0; i < candidates.size(); i++)
    {
      if (bestClauses.size() < numEstBestClauses_)
      {
        double candGain = candidates[i]->getAuxClauseData()->gain;
        double candAbsWt = fabs(candidates[i]->getWt());
        if (candGain > minGain_ && candAbsWt >= minWt_) 
          bestClauses.append(candidates[i]);
        else
        {
          //cout << "\tDiscarding candidate because of low gain or abs weight:";
          //cout << " gain = " << candGain << ", wt = " 
          //     << candidates[i]->getWt() << endl;
          //cout << "\t"; 
          //candidates[i]->printWithoutWtWithStrVar(cout, (*domains_)[0]);
          //cout << endl;
          delete candidates[i];
        }
      }
      else
        delete candidates[i];
    }

      //check whether the best clause has changed
    bool bestClauseChanged;
    if (bestClauses.size() > 0)
    {
      if (prevBestClause == NULL) bestClauseChanged = true;
      else
      {
          // if the prev and cur best clauses are the same 
        if (bestClauses[0]->same(prevBestClause) && 
            bestClauses[0]->getAuxClauseData()->op == 
            prevBestClause->getAuxClauseData()->op) 
          bestClauseChanged =  false;
        else
        { // prev and cur best clauses are different
            //if they have the same gain
          if (bestClauses[0]->getAuxClauseData()->gain >
              prevBestClause->getAuxClauseData()->gain) 
            bestClauseChanged = true;
          else
            bestClauseChanged = false;
        }
      }
    }
    else
      bestClauseChanged = false;

    if (prevBestClause) delete prevBestClause;
    return bestClauseChanged;
  }


  bool effectBestCandidateOnMLNs(Clause* const & cand, double& score)
  {
    if (appendToAndRemoveFromMLNs(cand, score))
    { 
      printMLNClausesWithWeightsAndScore(score, iter_); 
      //printMLNToFile(NULL, iter_);
      return true;
    }
    return false;
  }


  bool effectBestCandidateOnMLNs(Array<Clause*>& bestCandidates, double& score)
  {
    cout << "effecting best candidate on MLN..." << endl << endl;
    bool ok = false;
    int i;
    for (i = 0; i < bestCandidates.size(); i++)
    {
      cout << "effecting best candidate " << i << " on MLN..." << endl << endl;
      if (ok=effectBestCandidateOnMLNs(bestCandidates[i], score)) {
	cout << "!!!!!!!!!!Added Best Candidate\n";
	break;
      }
      cout << "failed to effect candidate on MLN." << endl;
      delete bestCandidates[i];
    }
    for (int j = i+1; j < bestCandidates.size();j++) delete bestCandidates[j];
    return ok;
  }

  //same as original method except that it returns the index of the
  //candidate that was added. Used to keep stuff found during
  //relational pathfinding out of the beam. Returns -1 if not ok
  int effectBestCandidateOnMLNsReturnIndex(Array<Clause*>& bestCandidates, double& score)
  {
    cout << "effecting best candidate on MLN..." << endl << endl;
    int index = -1;
    int i;
    for (i = 0; i < bestCandidates.size(); i++)
    {
      cout << "effecting best candidate " << i << " on MLN..." << endl << endl;
      if (effectBestCandidateOnMLNs(bestCandidates[i], score)) {
	cout << "Added Best Candidate\n";
	index = i;
	break;
      }
      cout << "failed to effect candidate on MLN." << endl;
      delete bestCandidates[i];
    }
    for (int j = i+1; j < bestCandidates.size();j++) delete bestCandidates[j];
    return index;
  }



  double getPenalty(const Clause* const & cand)
  {
    int op = cand->getAuxClauseData()->op;

    if (op == OP_ADD) return cand->getNumPredicates() * penalty_;

    if (op == OP_REPLACE_ADDPRED) 
    {
      int remIdx = cand->getAuxClauseData()->removedClauseIdx;
      int origlen = mln0_->getClause(remIdx)->getNumPredicates();        
      int diff = cand->getNumPredicates() - origlen;
      assert(diff > 0);
      return diff * penalty_;
    }

    if (op == OP_REPLACE_REMPRED) 
    {
      int remIdx = cand->getAuxClauseData()->removedClauseIdx;
      int origlen = mln0_->getClause(remIdx)->getNumPredicates();        
      int diff = origlen - cand->getNumPredicates();
      assert(diff > 0);
      return diff * penalty_;
    }

    if (op == OP_REPLACE)
    {
        //NOTE: op is REPLACE only when an MLN clause is to be replaced with
        //      one that is identical except for its predicates' senses
      int remIdx = cand->getAuxClauseData()->removedClauseIdx;
      const Clause* mlnClause = mln0_->getClause(remIdx);
      assert(cand->getNumPredicates() == mlnClause->getNumPredicates());
      int diff = 0;
      for (int i = 0; i < cand->getNumPredicates(); i++)
      {
        Predicate* cpred = (Predicate*) cand->getPredicate(i);
        Predicate* mpred = (Predicate*) mlnClause->getPredicate(i);
        assert(cpred->same(mpred));
        if (cpred->getSense() != mpred->getSense()) diff++;
      }
      return diff * penalty_;      
    }

    if (op == OP_REMOVE) return cand->getNumPredicates() * penalty_;

    assert(false);
    return 88888;
  }


  void printNewScore(const Array<Clause*>& carray, const Domain* const & domain,
                     const int& lbfgsbIter, const double& lbfgsbSec,
                     const double& newScore, const double& gain,
                     const Array<double>& penalties)
  {
    cout << "*************************** " << candCnt_++ << ", iter " << iter_ 
         << ", beam search iter " << bsiter_ << endl;
    for (int i = 0; i < carray.size(); i++)
    {
      if (carray[i]) 
      { 
        cout << "candidate     : ";
        carray[i]->printWithoutWtWithStrVar(cout,domain); cout << endl;
        cout << "op            : ";
        cout << Clause::getOpAsString(carray[i]->getAuxClauseData()->op) <<endl;
        cout << "removed clause: ";
        int remIdx = carray[i]->getAuxClauseData()->removedClauseIdx;
        if (remIdx < 0) cout << "NULL";
        else { mln0_->getClause(remIdx)->printWithoutWtWithStrVar(cout,domain);}
        cout << endl;
        if (carray[i]->getAuxClauseData()->prevClauseStr.length() > 0)
        {
          cout << "prevClause    : ";
          cout << carray[i]->getAuxClauseData()->prevClauseStr << endl;
        }
        if (carray[i]->getAuxClauseData()->addedPredStr.length() > 0)
        {
        cout << "addedPred     : ";
          cout << carray[i]->getAuxClauseData()->addedPredStr << endl;
        }
        if (carray[i]->getAuxClauseData()->removedPredIdx >= 0)
        {
          cout << "removedPredIdx: ";
          carray[i]->getAuxClauseData()->removedPredIdx;
        }

        cout << "score    : " << newScore << endl;
        cout << "gain     : " << gain << endl;
        cout << "penalty  : " << penalties[i] << endl;
        cout << "net gain : " << gain - penalties[i] << endl;
        if (carray[i]->getAuxClauseData()->op != OP_REMOVE)
          cout << "wt       : " << carray[i]->getWt() << endl;
      }
    }
    cout << "num LBFGSB iter      = " << lbfgsbIter << endl;
    cout << "time taken by LBFGSB = "; Timer::printTime(cout, lbfgsbSec);
    cout << endl;
    cout << "*************************** " << endl << endl;;
  }


  void printNewScore(const Clause* const & c, const Domain* const & domain, 
                     const int& lbfgsbIter, const double& lbfgsbSec,
                     const double& newScore, const double& gain,
                     const double& penalty)
  {
    Array<Clause*> carray;
    Array<double> parray;
    if (c) { carray.append((Clause*)c); parray.append(penalty); }
    printNewScore(carray, domain, lbfgsbIter, lbfgsbSec, newScore, gain,parray);
  }


  void printMLNClausesWithWeightsAndScore(const double& score, const int& iter)
  {    
    if (iter >= 0) cout << "MLN in iteration " << iter << ":" << endl;
    else           cout << "MLN:" << endl;
    cout << "------------------------------------" << endl;
    if (indexTrans_) indexTrans_->printClauseFormulaWts(cout, true);
    else             mln0_->printMLNClausesFormulas(cout, (*domains_)[0], true);
    cout << "------------------------------------" << endl;
    cout << "score = "<< score << endl << endl;
  }


  void printMLNToFile(const char* const & appendStr, const int& iter)
  {
    string fname = outMLNFileName_;

    if (appendStr) fname.append(".").append(appendStr);

    if (iter >= -1) 
    { 
      char buf[100]; 
      sprintf(buf, "%d", iter);
      fname.append(".iter").append(buf);
    }

    if (appendStr || iter >= -1) fname.append(".mln");
    
    ofstream out(fname.c_str());
    if (!out.good()) { cout << "ERROR: failed to open " <<fname<<endl;exit(-1);}
        // output the predicate declaration
    out << "//predicate declarations" << endl;
    (*domains_)[0]->printPredicateTemplates(out);
    out << endl;

    if (indexTrans_) indexTrans_->printClauseFormulaWts(out, false);
    else             mln0_->printMLNClausesFormulas(out, (*domains_)[0], false);

    out << endl;
    out.close();
  } 
  

  void printClausesInBeam(const ClauseOpHashArray* const & beam)
  {
    cout.setf(ios_base::left, ios_base::adjustfield);
    cout << "^^^^^^^^^^^^^^^^^^^ beam ^^^^^^^^^^^^^^^^^^^" << endl;
    for (int i = 0; i < beam->size(); i++)
    {
      cout << i << ":  ";
      cout.width(10); cout << (*beam)[i]->getWt(); cout.width(0); cout << " ";
      (*beam)[i]->printWithoutWtWithStrVar(cout, (*domains_)[0]); cout << endl;
    }
    cout.width(0);
    cout << "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^" << endl;
    
  }

  void setPriorMeansStdDevs(Array<double>& priorMeans, 
                            Array<double>& priorStdDevs, const int& addSlots,
                            const Array<int>* const& removedSlotIndexes)
  {
    priorMeans.clear(); priorStdDevs.clear();

    if (indexTrans_)
    {
      indexTrans_->setPriorMeans(priorMeans);
      priorStdDevs.growToSize(priorMeans.size());
      for (int i = 0; i < priorMeans.size(); i++)
        priorStdDevs[i] = priorStdDev_;
    }
    else
    {
      for (int i = 0; i < mln0_->getNumClauses(); i++)
      {
        priorMeans.append(mln0_->getMLNClauseInfo(i)->priorMean);
        priorStdDevs.append(priorStdDev_);
      }
    }

    if (removedSlotIndexes)
    {
      for (int i = 0; i < removedSlotIndexes->size(); i++)
        priorMeans[ (*removedSlotIndexes)[i] ] = 0;
    }

    for (int i = 0; i < addSlots; i++)
    {
      priorMeans.append(priorMean_);
      priorStdDevs.append(priorStdDev_);      
    }
  }


  void pllSetPriorMeansStdDevs(Array<double>& priorMeans, 
                               Array<double>& priorStdDevs, const int& addSlots,
                               const Array<int>* const & removedSlotIndexes)
  {
    if (hasPrior_)
    {
      setPriorMeansStdDevs(priorMeans, priorStdDevs, 
                           addSlots, removedSlotIndexes);
      pll_->setMeansStdDevs(priorMeans.size(), priorMeans.getItems(), 
                            priorStdDevs.getItems());
    }
    else
      pll_->setMeansStdDevs(-1, NULL, NULL);
  }


  void setRemoveAppendedPriorMeans(const int& numClausesFormulas,
                                   Array<double>& priorMeans, 
                                   const Array<int>& removedSlotIndexes, 
                                   const int& addSlots,
                                   Array<double>& removedValues)
  {
    for (int i = 0; i < removedSlotIndexes.size(); i++)
    {
      removedValues.append(priorMeans[ removedSlotIndexes[i] ]);
      priorMeans[ removedSlotIndexes[i] ] = 0;
    }

    for (int i = 0; i < addSlots; i++)
      priorMeans[numClausesFormulas+i] = priorMean_;
  }

  
  int getNumClausesFormulas()
  {
    if (indexTrans_) return indexTrans_->getNumClausesAndExistFormulas();
    return mln0_->getNumClauses();
  }


  void updateWts(const Array<double>& wts,
                 const Array<Clause*>* const & appendedClauses,
                 const Array<string>* const & appendedFormulas)
  {
    if (indexTrans_)
    {
      Array<double> tmpWts; 
      tmpWts.growToSize(wts.size()-1);
      for (int i = 1; i < wts.size(); i++) tmpWts[i-1] = wts[i];
      indexTrans_->updateClauseFormulaWtsInMLNs(tmpWts, appendedClauses,
                                                appendedFormulas);
    }
    else
    {
      for (int i = 0; i < mlns_->size(); i++)
        for (int j = 0; j < (*mlns_)[i]->getNumClauses(); j++)          
          ((Clause*)(*mlns_)[i]->getClause(j))->setWt(wts[j+1]);
    }
  }


  /////////////////////////// quicksort ////////////////////////////////

 private:
    //sort clauses in decreasing order of gain
  void rquicksort(Array<Clause*>& clauses, const int& l, const int& r)  
  {
    Clause** items = (Clause**) clauses.getItems();
    if (l >= r) return;
    Clause* tmp = items[l];
    items[l] = items[(l+r)/2];
    items[(l+r)/2] = tmp;

    int last = l;
    for (int i = l+1; i <= r; i++)
      if (items[i]->getAuxClauseData()->gain > 
          items[l]->getAuxClauseData()->gain)
      {
        ++last;
        Clause* tmp = items[last];
        items[last] = items[i];
        items[i] = tmp;
      }
    
    tmp = items[l];
    items[l] = items[last];
    items[last] = tmp;
    rquicksort(clauses, l, last-1);
    rquicksort(clauses, last+1, r); 
  }

  void rquicksort(Array<Clause*>& ca) { rquicksort(ca, 0, ca.size()-1); }



  ////////////////// functions to handle existential formulas ///////////////
 private:
  void deleteExistFormulas(Array<ExistFormula*>& existFormulas);

  void getMLNClauses(Array<Clause*>& initialMLNClauses,
                     Array<ExistFormula*>& existFormulas);

    //cnfClauses enter & exit function with its AuxClauseData's cache == NULL
  inline void pllCountsForExistFormula(Clause* cnfClause, const int& domainIdx,
                                       int* clauseIdxInMLN,
                                       Array<UndoInfo*>* const & undoInfos);

  inline void printNewScore(const string existFormStr, const int& lbfgsbIter, 
                            const double& lbfgsbSec, const double& newScore, 
                            const double& gain, const double& penalty, 
                            const double& wt);

  inline void rquicksort(Array<ExistFormula*>& existFormulas, const int& l, 
                         const int& r);

  inline void rquicksort(Array<ExistFormula*>& ef);
 
  inline void evaluateExistFormula(ExistFormula* const& ef,
                                   const bool& computeCountsOnly,
              Array<Array<Array<IndexAndCount*> > >* const & iacArraysPerDomain,
                                   const double& prevScore);

  double evaluateExistFormulas(Array<ExistFormula*>& existFormulas, 
                               Array<ExistFormula*>& highGainWtFormulas,
                               const double& prevScore);

  inline void appendExistFormulaToMLNs(ExistFormula* const & ef);


  inline bool effectExistFormulaOnMLNs(ExistFormula* ef, 
                                       Array<ExistFormula*>& existFormulas, 
                                       double& score);

  bool effectBestCandidateOnMLNs(Array<Clause*>& bestCandidates, 
                                 Array<ExistFormula*>& existFormulas,
                                 Array<ExistFormula*>& highGainWtFormulas,
                                 double& score);
  /////////////////////////////////////////////////////////////////////////
 private:
    // mln0_ and mlns_ are not owned by StructLearn; do not delete;
    // if there are more than one domains, mln0_ corresponds to domains_[0]
  MLN* mln0_; 
  Array<MLN*>* mlns_;
  bool startFromEmptyMLN_;
  string outMLNFileName_;
  Array<Domain*>* domains_; // not owned by StructLearn; do not delete;
  Array<Predicate*>* preds_;
  Array<bool>* areNonEvidPreds_;
  ClauseFactory* clauseFactory_;

  bool cacheClauses_;
  bool origCacheClauses_;
  ClauseHashArray* cachedClauses_;
  double cacheSizeMB_;
  double maxCacheSizeMB_;

  bool tryAllFlips_;
  bool sampleClauses_;
  bool origSampleClauses_;

  PseudoLogLikelihood* pll_;
  bool hasPrior_;
  double priorMean_;
  double priorStdDev_;
  bool wtPredsEqually_;

  LBFGSB* lbfgsb_;
  int lbMaxIter_; 
  double lbConvThresh_;
  int looseMaxIter_; 
  double looseConvThresh_;

  int beamSize_;
  int bestGainUnchangedLimit_;
  int numEstBestClauses_;
  double minGain_;
  double minWt_;
  double penalty_;

  bool sampleGndPreds_;
  double fraction_;
  int minGndPredSamples_;
  int maxGndPredSamples_;

  bool reEvalBestCandsWithTightParams_;

  Timer timer_;
  int candCnt_;

  int iter_;
  int bsiter_;
  double startSec_;

    //Used to map clause indexes in MLNs to the weights indexes that are 
    //presented to LBFGSB. This is required if the clauses do not line up
    //across databases (as when an existentially quantified formula has a 
    //different number of formulas in its CNF for different databases).
  IndexTranslator* indexTrans_;
};
#endif
