/*
 * Author: Tuyen Huynh
 * Description: Find the best predicate mapping between the source and target domain.
 */

#include <fstream>
#include <iostream>
#include "arguments.h"
#include "predicatesMapping.h"

char* sourceMLNFile = NULL;
char* targetPredicateFile = NULL;
char* targetDbFiles = NULL;
char* mappingConstraintFile = NULL;
char* outTargetMLNFile = NULL;
bool  globalMapping = false;
bool  doRevision = false;
char* funcFiles = NULL;
bool  sampling = false;
bool  debug = false;
double confidentRatio = 0.01;

ARGS ARGS::Args[] = 
{
  ARGS("sourceMLN", ARGS::Req, sourceMLNFile, 
       "Source .mln file containing the source Markov Logic Network. "),

  ARGS("targetPredicate", ARGS::Req, targetPredicateFile, 
       "Target predicates description file. "),

  ARGS("targetDb", ARGS::Req, targetDbFiles, 
       "Comma-separated .db files containinsg the target training database "
       "(of true/false ground atoms), including function definitions, "
       "e.g. ai.db,graphics.db,languages.db."), 
       
  ARGS("targetMLNOut", ARGS::Req, outTargetMLNFile, 
       "Filename of the transfered target MLN "),
 
  ARGS("constraint", ARGS::Opt, mappingConstraintFile, 
       "Type constraint description file. "),

  ARGS("sampling", ARGS::Opt, sampling,
  	   "Perform sampling instead of exact counting"),
  
  ARGS("globalMapping", ARGS::Opt, globalMapping,
  	   "Perform global mapping, if not provided perform local mapping"),
  	            
  ARGS("doRevision", ARGS::Opt, doRevision,
  	   "Revising the clause after do the mapping, if not provided perform the mapping"),
  	   
  ARGS("confidentRatio", ARGS::Opt, confidentRatio,
   	   "default value is 0.01"),
  	            
  ARGS("debug", ARGS::Opt, debug,
  	   "Print out debug information"),
  
  ARGS()
};

int main(int argc, char* argv[])
{
  ARGS::parse(argc,argv,&cout);
  Timer timer;
  double startSec = timer.time();
  
  PredicatesMapping predicatesMapping(sourceMLNFile, targetPredicateFile, targetDbFiles, 
  										mappingConstraintFile, outTargetMLNFile, globalMapping, sampling, debug, doRevision, confidentRatio);
  predicatesMapping.run();
  cout << "Total time taken = "; 
  Timer::printTime(cout, timer.time()-startSec); cout << endl;
}
