#!/usr/bin/perl 
#
# Copyright 2006, Tuyen Huynh
#
use strict;
use warnings;
use Shell;
use File::Copy;
use constant USAGE =>
"usage: ./srcMLNToTargetMLN.pl <sourceMLNFile> <predicateMappingFile> <outputFile>";

# declare global variables
use vars qw (@spreds @tmp_tpreds @tpreds $srcMLN $predMapping $output);

# Declare the subroutines
sub trim($);
@ARGV == 3 or die USAGE;
my ( $srcMLN, $predMapping, $output ) = @ARGV;
### main program ###

# read predicate mapping file
open( MAPPING, $predMapping );
my @lines = <MAPPING>;
close(MAPPING);
chomp foreach @lines;
my $index = 0;
for my $i ( 0 .. $#lines ) {
	my $aline = $lines[$i];
	( $spreds[$index], $tpreds[$index] ) = split( /->/, $aline );
	$spreds[$index] = trim( $spreds[$index] );
	$tpreds[$index] = trim( $tpreds[$index] );
	$tmp_tpreds[$index] = $tpreds[$index];
	$tmp_tpreds[$index] =~ tr/a-z/A-Z/;
	$index          = $index + 1;
}

# if the mapping is empty, producing empty network
if (!(@spreds)) {
	open( OUT, ">", $output) or die "Cannot open file: $!";
	close(OUT) or die $!;
	exit;
}

# read source MLN
open(SRCMLN, $srcMLN) or die $!;
my @srcLines = <SRCMLN>;
close(SRCMLN);
chomp foreach @srcLines;

open( OUT, ">", $output) or die "Cannot open file: $!";
for my $i ( 0 .. $#srcLines ) {
	my $aline = $srcLines[$i];

	# translate into internal representation
	for my $j ( 0 .. $#spreds ) {
		if ($aline ne '') {
			my $old = $spreds[$j];
			my $new = $tmp_tpreds[$j ];
			if (($new eq '') && ($aline =~ /$old/))
			{
				$aline = '';
			}else {
				$aline =~ s/$old/$new/g;
			}
		}
	}

	if ($aline ne '') {
		#translate into target domain
		for my $k ( 0 .. $#tpreds ) {
			my $old = $tmp_tpreds[$k];
			my $new = $tpreds[$k];
			$aline =~ s/$old/$new/g;
		}
		print OUT "$aline\n";
	}
}
close(OUT) or die $!;
exit;

# Perl trim function to remove whitespace from the start and end of the string
sub trim($) {
	my $string = shift;
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}

